
/*******************************************************************************
*
* File check7.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the assignment programs assign_swd2swbgr, ...
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "sw_term.h"
#include "global.h"


static void random_swd(void)
{
   int k;
   pauli_dble *pa,*pam;

   pa=swd;
   pam=swd+2*VOLUME;

   for (;pa<pam;pa++)
   {
      ranlxd((*pa).u,36);

      for (k=0;k<36;k++)
      {
         if (k<6)
            (*pa).u[k]=1.0+0.05*(0.5-(*pa).u[k]);
         else
            (*pa).u[k]=0.05*(0.5-(*pa).u[k]);
      }
   }

   set_flags(COMPUTED_SWD);
}


static int cmp_sw(pauli *r,pauli *s)
{
   float *rr,*rs,*rm;

   rr=(float*)((*r).u);
   rs=(float*)((*s).u);
   rm=rr+72;

   for (;rr<rm;rr++)
   {
      if ((*rr)!=(*rs))
         return 1;
      rs+=1;
   }

   return 0;
}


static int cmp_swd(pauli_dble *r,pauli_dble *s)
{
   double *rr,*rs,*rm;

   rr=(double*)((*r).u);
   rs=(double*)((*s).u);
   rm=rr+72;

   for (;rr<rm;rr++)
   {
      if ((*rr)!=(*rs))
         return 1;
      rs+=1;
   }

   return 0;
}


static int check_sw(block_t *b)
{
   int x0,x1,x2,x3,x[4];
   int y0,y1,y2,y3;
   int ix,iy;

   for (x0=0;x0<(*b).bs[0];x0++)
   {
      for (x1=0;x1<(*b).bs[1];x1++)
      {
         for (x2=0;x2<(*b).bs[2];x2++)
         {
            for (x3=0;x3<(*b).bs[3];x3++)
            {
               x[0]=x0;
               x[1]=x1;
               x[2]=x2;
               x[3]=x3;

               y0=(*b).bo[0]+x0;
               y1=(*b).bo[1]+x1;
               y2=(*b).bo[2]+x2;
               y3=(*b).bo[3]+x3;

               ix=ipt_blk(b,x);
               iy=ipt[y3+L3*y2+L2*L3*y1+L1*L2*L3*y0];

               if (cmp_sw(&(*b).sw[2*ix],&sw[2*iy])!=0)
                     return 1;
            }
         }
      }
   }

   return 0;
}


static int check_swd(block_t *b)
{
   int x0,x1,x2,x3,x[4];
   int y0,y1,y2,y3;
   int ix,iy;

   for (x0=0;x0<(*b).bs[0];x0++)
   {
      for (x1=0;x1<(*b).bs[1];x1++)
      {
         for (x2=0;x2<(*b).bs[2];x2++)
         {
            for (x3=0;x3<(*b).bs[3];x3++)
            {
               x[0]=x0;
               x[1]=x1;
               x[2]=x2;
               x[3]=x3;

               y0=(*b).bo[0]+x0;
               y1=(*b).bo[1]+x1;
               y2=(*b).bo[2]+x2;
               y3=(*b).bo[3]+x3;

               ix=ipt_blk(b,x);
               iy=ipt[y3+L3*y2+L2*L3*y1+L1*L2*L3*y0];

               if (cmp_swd(&(*b).swd[2*ix],&swd[2*iy])!=0)
                  return 1;
            }
         }
      }
   }

   return 0;
}


static int check_swbgr(blk_grid_t grid)
{
   int nb,isw;
   block_t *b,*bm;

   b=blk_list(grid,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      if (check_sw(b)!=0)
         return 1;
   }

   return 0;
}


static int check_swdbgr(blk_grid_t grid)
{
   int nb,isw;
   block_t *b,*bm;

   b=blk_list(grid,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      if (check_swd(b)!=0)
         return 1;
   }

   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,n,nb,isw,ifail;
   int bs[4];
   ptset_t set;
   block_t *b;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check7.log","w",stdout);
      fin=freopen("check7.in","r",stdin);

      printf("\n");
      printf("Check of the assignment programs assign_swd2swbgr, ...\n");
      printf("------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_sw();
   alloc_swd();
   alloc_bgr(GCR_BLOCKS,bs);
   alloc_bgr(HMC_BLOCKS,bs);

   ifail=0;

   for (set=0;set<=NO_PTS;set++)
   {
      random_swd();
      ifail+=assign_swd2swbgr(GCR_BLOCKS,set);
      ifail+=invert_swd(set);
      assign_swd2sw();
      error(check_swbgr(GCR_BLOCKS)!=0,1,
            "main [check7.c]","assign_swd2swbgr is incorrect");

      random_swd();
      ifail+=assign_swd2swdbgr(HMC_BLOCKS,set);
      ifail+=invert_swd(set);
      error(check_swdbgr(HMC_BLOCKS)!=0,1,
            "main [check7.c]","assign_swd2swdbgr is incorrect");

      random_swd();
      ifail+=assign_swd2swdbgr(HMC_BLOCKS,NO_PTS);
      ifail+=invert_swd(set);
      assign_swd2sw();
      b=blk_list(HMC_BLOCKS,&nb,&isw);

      for (n=0;n<nb;n++)
      {
         ifail+=assign_swdb2swb(HMC_BLOCKS,n,set);
         error(check_sw(b)!=0,1,
               "main [check7.c]","assign_swdb2swb is incorrect");
         b+=1;
      }
   }

   error_chk();
   error(ifail!=0,1,
         "main [check7.c]","Some of the inversions were not safe");

   if (my_rank==0)
   {
      printf("\n");
      printf("No errors detected --- all programs work correctly\n\n");
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
