
/*******************************************************************************
*
* File check7.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Direct check of Znohat_dble()
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "dfl.h"
#include "global.h"


static void random_basis(int Ns)
{
   int i;
   spinor **ws;

   ws=reserve_ws(Ns);

   for (i=0;i<Ns;i++)
      random_s(VOLUME,ws[i],1.0f);
   
   dfl_subspace(ws);

   release_ws();
}


int main(int argc,char *argv[])
{
   int my_rank,ic;
   int bs[4],Ns,nb,nv;
   int iwd0,iwd1,iwd2;
   double d,dmax;
   complex_dble **wvd,spv,sps;
   spinor_dble **wsd;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check7.log","w",stdout);
      fin=freopen("check7.in","r",stdin);
      
      printf("\n");
      printf("Direct check of Znohat_dble()\n");
      printf("-----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("Ns","%d",&Ns);
      fclose(fin);

      if (Ns<4)
         Ns=4;
      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("Ns = %d\n\n",Ns);
      fflush(flog);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,123456);   
   geometry();

   alloc_ud();
   alloc_s(Ns);
   alloc_sd(3);

   set_lat_parms(0.0,0.1,1.0);
   set_dfl_parms(bs,Ns,4,1,1,1,1,1,1,0.12,1.0e-3,1.0e-6,1.0e-12,0,4);
   
   nb=VOLUME/(bs[0]*bs[1]*bs[2]*bs[3]);
   nv=nb*Ns;

   wsd=reserve_wsd(3);
   wvd=reserve_wvd(3);
   iwd0=(wsd[0]-psd[0][0])/NSPIN;
   iwd1=(wsd[1]-psd[0][0])/NSPIN;
   iwd2=(wsd[2]-psd[0][0])/NSPIN;
   dmax=0.0;

   for (ic=0;ic<4;ic++)
   {
      if ((ic==0)||(ic==3))
         random_ud();
      if (ic==2)
         flipbcd();
      if (ic!=1)
         random_basis(Ns);      
      random_vd(nv,wvd[0],1.0);
      random_vd(nv,wvd[1],1.0);

      Znohat_dble(wvd[0],wvd[2]);
      spv=vprod_dble(nv,1,wvd[1],wvd[2]);

      dfl_vd2sd(wvd[0],iwd0);
      dfl_vd2sd(wvd[1],iwd1);

      Qnohat_dble(iwd0,iwd2);
      mulg5_dble(VOLUME,wsd[2]);
      sps=spinor_prod_dble(VOLUME,1,wsd[1],wsd[2]);
      sps.re-=spv.re;
      sps.im-=spv.im;
      d=sps.re*sps.re+sps.im*sps.im;
      d/=(vnorm_square_dble(nv,1,wvd[1])*vnorm_square_dble(nv,1,wvd[2]));

      if (d>dmax)
         dmax=d;
   }

   error_chk();
   message("Maximal relative deviation = %.1e\n\n",sqrt(dmax));
   
   if (my_rank==0)
      fclose(flog);
   
   MPI_Finalize();   
   exit(0);
}
