
/*******************************************************************************
*
* File check9.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generation of the deflation subspace
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "flags.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,id;
   int bc,nc_first,nc_last;
   int bs_sap[4],nkv,nmr,ncy,nmx;
   int bs_dfl[4],Ns,rinv,rnmr,rncy,rvnmx;
   int vnkv,vnmx,vdnmx;
   int nsize,icnfg,n;
   int iw0,iw1,iwd0,iwd1,status,ifail,stat[3];
   double beta,kappa,csw,res,r;
   double rkappa,rvres,vres,vdres;
   complex **v;
   spinor **ws;
   spinor_dble **wsd;
   char cnfg_dir[NAME_SIZE],cnfg_file[NAME_SIZE];
   char nbase[NAME_SIZE];   
   dfl_parms_t dfl;
   FILE *fin=NULL,*flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check9.log","w",stdout);
      fin=freopen("check9.in","r",stdin);
      
      printf("\n");
      printf("Generation of the deflation subspace\n");
      printf("------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("id","%d\n",&id);
      read_line("cnfg_dir","%s\n",cnfg_dir);
      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);      
      read_line("bc","%d\n",&bc);      
      read_line("nc_first","%d\n",&nc_first);
      read_line("nc_last","%d\n",&nc_last);      

      read_line("bs_sap","%d %d %d %d\n",
                &bs_sap[0],&bs_sap[1],&bs_sap[2],&bs_sap[3]);
      read_line("nkv","%d\n",&nkv);      
      read_line("nmr","%d\n",&nmr);
      read_line("ncy","%d\n",&ncy);      
      read_line("nmx","%d\n",&nmx);      
      read_line("res","%lf\n",&res);
      
      read_line("bs_dfl","%d %d %d %d\n",
                &bs_dfl[0],&bs_dfl[1],&bs_dfl[2],&bs_dfl[3]);
      read_line("Ns","%d\n",&Ns);
      read_line("rkappa","%lf\n",&rkappa);
      read_line("rinv","%d\n",&rinv);
      read_line("rnmr","%d\n",&rnmr);
      read_line("rncy","%d\n",&rncy);
      read_line("rvnmx","%d\n",&rvnmx);      
      read_line("rvres","%lf\n",&rvres);
      read_line("vnkv","%d\n",&vnkv);
      read_line("vnmx","%d\n",&vnmx);
      read_line("vres","%lf\n",&vres);
      read_line("vdnmx","%d\n",&vdnmx);
      read_line("vdres","%lf",&vdres);
 
      fclose(fin);

      printf("beta = %.6f\n",beta);
      printf("kappa = %.6f\n",kappa);
      printf("csw = %.6f\n",csw);
      printf("bc = %+d\n",bc);

      printf("bs_sap = %d %d %d %d\n",
             bs_sap[0],bs_sap[1],bs_sap[2],bs_sap[3]);
      printf("nkv = %d\n",nkv);
      printf("nmr = %d\n",nmr);      
      printf("ncy = %d\n",ncy);
      printf("nmx = %d\n",nmx);
      printf("res = %.1e\n",res);       
      
      printf("bs_dfl = %d %d %d %d\n",
             bs_dfl[0],bs_dfl[1],bs_dfl[2],bs_dfl[3]);
      printf("Ns = %d\n",Ns);
      printf("rkappa = %.6f\n",rkappa);       
      printf("rinv = %d\n",rinv);
      printf("rnmr = %d\n",rnmr);       
      printf("rncy = %d\n",rncy);
      printf("rvnmx = %d\n",rvnmx);
      printf("rvres = %.1e\n",rvres);       
      printf("vnkv = %d\n",vnkv);
      printf("vnmx = %d\n",vnmx);
      printf("vres = %.1e\n",vres);
      printf("vdnmx = %d\n",vdnmx);
      printf("vdres = %.1e\n\n",vdres);
     
      printf("Configurations no %d -> %d\n\n",nc_first,nc_last);
      fflush(flog);      
   }

   MPI_Bcast(&id,1,MPI_INT,0,MPI_COMM_WORLD);      
   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_first,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nc_last,1,MPI_INT,0,MPI_COMM_WORLD);   

   MPI_Bcast(bs_sap,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   MPI_Bcast(bs_dfl,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&Ns,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rkappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&rinv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rnmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rncy,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&rvnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&rvres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vnkv,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&vnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdnmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&vdres,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   start_ranlux(0,123456);   
   geometry();

   alloc_u();
   alloc_ud();
   alloc_sw();
   alloc_swd();

   if (Ns<(2*nkv+2))
      alloc_s(2*nkv+3);
   else
      alloc_s(Ns+2);
   alloc_sd(6);

   v=amalloc(Ns*sizeof(*v),3);
   error(v==NULL,1,"main [check9.c]","Unable to allocate auxiliary array");
   
   set_gcr_parms(bs_sap,nmr,ncy,nkv);
   dfl=set_dfl_parms(bs_dfl,Ns,rinv,rnmr,rncy,rvnmx,vnkv,vnmx,vdnmx,
                     rkappa,rvres,vres,vdres,0,0);

   error((dfl.Ns!=Ns)||(dfl.rinv!=rinv)||(dfl.rnmr!=rnmr)||
         (dfl.rncy!=rncy)||(dfl.rvnmx!=rvnmx)||(dfl.vnkv!=vnkv)||
         (dfl.vnmx!=vnmx)||(dfl.vdnmx!=vdnmx)||(dfl.rkappa!=rkappa)||
         (dfl.rvres!=rvres)||(dfl.vres!=vres)||(dfl.vdres!=vdres),1,
         "main [check9.c]","dfl parameters are incorrectly set");
   
   check_dir_root(cnfg_dir);   

   nsize=name_size("%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
                   NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);
   error_root(nsize>=NAME_SIZE,1,"main [check9.c]","Base name is too long");
   sprintf(nbase,"%dx%dx%dx%db%.2fk%.5fc%.5fid%d",
           NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3,beta,kappa,csw,id);

   nsize=name_size("%s/%sn%d",cnfg_dir,nbase,nc_last);
   error_root(nsize>=NAME_SIZE,1,"main [check9.c]",
              "cnfg_dir name is too long");
    
   for (icnfg=nc_first;icnfg<=nc_last;icnfg++)
   {
      if (my_rank==0)
      {
         printf("Configuration no %d:\n\n",icnfg);
         fflush(flog);
      }

      set_lat_parms(beta,kappa,csw);
      sprintf(cnfg_file,"%s/%sn%d",cnfg_dir,nbase,icnfg);
      import_cnfg(cnfg_file);
      sw_term();
      if (bc==-1)
         flipbcd();

      dfl_modes(1,&status,&ifail);

      if (my_rank==0)
         fflush(flog);
      
      dfl_global_modes(v);

      ws=reserve_ws(2);
      iw0=(ws[0]-ps[0][0])/NSPIN;
      iw1=(ws[1]-ps[0][0])/NSPIN;
      
      wsd=reserve_wsd(2);
      iwd0=(wsd[0]-psd[0][0])/NSPIN;
      iwd1=(wsd[1]-psd[0][0])/NSPIN;

      message("Solution of the Dirac equation at kappa = %.6f:\n\n",kappa);
      
      for (n=0;n<Ns;n++)
      {
         dfl_v2s(v[n],iw0);
         Qnohat(iw0,iw1);
         r=(double)(norm_square(VOLUME,1,ws[1]));
         message("n = %2d, <Q^2> = %.1e\n",n,r);
      }

      random_sd(VOLUME,wsd[0],1.0);

      sap_gcr(nmx,res,iw0,iw1,stat,&ifail);
      message("\nSAP+GCR  solver iteration number (res = %.1e): %2d\n",
              res,stat[0]);
      if (my_rank==0)
         fflush(flog);

      dfl_sap_gcr(nmx,res,iw0,iw1,stat,&ifail);
      message("Deflated solver iteration number (res = %.1e): %2d "
              "(little: %2d, %2d)\n\n",res,stat[0],stat[1],stat[2]);
      if (my_rank==0)
         fflush(flog);

      error_chk();

      if (rkappa!=kappa)
      {
         message("Solution of the Dirac equation at kappa = %.6f:\n\n",
                 rkappa);

         set_lat_parms(beta,rkappa,csw);
         sw_term();
         assign_swd2sw();      
      
         for (n=0;n<Ns;n++)
         {
            dfl_v2s(v[n],iw0);
            Qnohat(iw0,iw1);
            r=(double)(norm_square(VOLUME,1,ws[1]));
            message("n = %2d, <Q^2> = %.1e\n",n,r);
         }

         random_sd(VOLUME,wsd[0],1.0);

         sap_gcr(nmx,res,iw0,iw1,stat,&ifail);
         message("\nSAP+GCR  solver iteration number (res = %.1e): %2d\n",
                 res,stat[0]);
         if (my_rank==0)
            fflush(flog);

         dfl_sap_gcr(nmx,res,iw0,iw1,stat,&ifail);

         message("Deflated solver iteration number (res = %.1e): %2d "
                 "(little: %2d, %2d)\n\n",res,stat[0],stat[1],stat[2]);
         if (my_rank==0)
            fflush(flog);
      }

      release_ws();
      release_wsd();
      error_chk();
   }
      
   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }
   
   MPI_Finalize();   
   exit(0);
}
