
/*******************************************************************************
*
* File check8b.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Comparision of Qoe_blk_dble, ... with Qoe_dble, ...
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,n0,n,nb,isw,bs[4];
   double d,doe,deo,dno,dht;
   complex_dble z;
   spinor_dble *ps0e,*ps1e,*ps2e;
   spinor_dble *ps0o,*ps1o,*ps2o;
   block_t *b0,*b;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check8b.log","w",stdout);
      fin=freopen("check8b.in","r",stdin);

      printf("\n");
      printf("Comparision of Qoe_blk_dble, ... with Qoe_dble, ...\n");
      printf("---------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_sd(2);
   alloc_swd();

   random_ud();
   set_lat_parms(0.0,0.1,0.2);
   sw_term();

   alloc_bgr(HMC_BLOCKS,bs);
   assign_ud2udbgr(HMC_BLOCKS);
   assign_swd2swdbgr(HMC_BLOCKS,NO_PTS);

   z.re=-1.0;
   z.im=0.0;
   doe=0.0;
   deo=0.0;
   dno=0.0;
   dht=0.0;

   b0=blk_list(HMC_BLOCKS,&nb,&isw);

   for (n0=0;n0<nb;n0++)
   {
      if (isw==0)
         n=n0;
      else
         n=(n0+(nb/2))%nb;
      b=b0+n;

      ps0e=&(*b).sd[0][0];
      ps1e=&(*b).sd[1][0];
      ps2e=&(*b).sd[2][0];

      ps0o=&(*b).sd[0][(*b).volh];
      ps1o=&(*b).sd[1][(*b).volh];
      ps2o=&(*b).sd[2][(*b).volh];

      set_sd2zero(VOLUME,psd[0][0]);
      random_sd((*b).vol,ps0e,1.0);
      assign_sdb2sd(b,EVEN_PTS,0,0);

      error(check_blk_flags(HMC_BLOCKS,QOE_DBLE,n)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qoe_blk_dble]");
      Qoe_blk_dble(b,0,1);
      error(check_flags(QOE_DBLE)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qoe_dble]");
      Qoe_dble(0,1);
      assign_sd2sdb(b,ODD_PTS,1,2);

      mulc_spinor_add_dble((*b).volh,ps2o,ps1o,z);
      d=norm_square_dble((*b).volh,1,ps2o)/norm_square_dble((*b).volh,1,ps0o);
      if (d>doe)
         doe=d;

      random_sd((*b).vol,ps0e,1.0);
      random_sd((*b).vol,ps1e,1.0);
      assign_sdb2sd(b,ODD_PTS,0,0);
      assign_sdb2sd(b,EVEN_PTS,1,1);

      error(check_blk_flags(HMC_BLOCKS,QEO_DBLE,n)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qeo_blk_dble]");
      Qeo_blk_dble(b,0,1);
      error(check_flags(QEO_DBLE)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qeo_dble]");
      Qeo_dble(0,1);
      assign_sd2sdb(b,EVEN_PTS,1,2);

      mulc_spinor_add_dble((*b).volh,ps2e,ps1e,z);
      d=norm_square_dble((*b).volh,1,ps2e)/norm_square_dble((*b).volh,1,ps0e);
      if (d>deo)
         deo=d;

      random_sd((*b).vol,ps0e,1.0);
      assign_sdb2sd(b,ALL_PTS,0,0);

      error(check_blk_flags(HMC_BLOCKS,QNOHAT_DBLE,n)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qnohat_blk_dble]");
      Qnohat_blk_dble(b,0,1);
      error(check_flags(QNOHAT_DBLE)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qnohat_dble]");
      Qnohat_dble(0,1);
      assign_sd2sdb(b,ALL_PTS,1,2);

      mulc_spinor_add_dble((*b).vol,ps2e,ps1e,z);
      d=norm_square_dble((*b).vol,1,ps2e)/norm_square_dble((*b).vol,1,ps0e);
      if (d>dno)
         dno=d;
   }

   error(assign_swd2swdbgr(HMC_BLOCKS,ODD_PTS)!=0,1,"main [check8b.c]",
         "Inversion of swd on the odd sites was not safe");

   for (n0=0;n0<nb;n0++)
   {
      if (isw==0)
         n=n0;
      else
         n=(n0+(nb/2))%nb;
      b=b0+n;

      ps0e=&(*b).sd[0][0];
      ps1e=&(*b).sd[1][0];
      ps2e=&(*b).sd[2][0];
      ps1o=&(*b).sd[1][(*b).volh];

      random_sd((*b).vol,ps0e,1.0);
      assign_sd2sd((*b).volh,ps0e,ps2e);
      apply_sw_dble((*b).volh,(*b).swd,ps2e,ps2e);
      Qoe_blk_dble(b,0,1);
      apply_sw_dble((*b).volh,(*b).swd+(*b).vol,ps1o,ps1o);
      Qeo_blk_dble(b,1,2);
      error(check_blk_flags(HMC_BLOCKS,QHAT_DBLE,n)!=0,1,"main [check8b.c]",
            "Fields are not in proper condition [Qhat_blk_dble]");
      Qhat_blk_dble(b,0,1);

      mulc_spinor_add_dble((*b).volh,ps1e,ps2e,z);
      d=norm_square_dble((*b).volh,1,ps1e)/norm_square_dble((*b).volh,1,ps0e);
      if (d>dht)
         dht=d;
   }

   error_chk();

   if (my_rank==0)
   {
      printf("|[Qoe_blk-Qoe]*psi|/|psi| = %.2e\n\n",
             sqrt(doe));

      printf("|[Qeo_blk-Qeo]*psi|/|psi| = %.2e\n\n",
             sqrt(deo));

      printf("|[Qnohat_blk-Qnohat]*psi|/|psi| = %.2e\n\n",
             sqrt(dno));

      printf("|[Qhat_blk-(Qee_blk-Qeo_blk*Qooinv_blk*Qoe_blk)]*psi|/|psi| ");
      printf("= %.2e\n\n",sqrt(dht));

      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
