
/*******************************************************************************
*
* File check16.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the block shift generator
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "start.h"
#include "misc.h"
#include "hmc.h"
#include "global.h"

static int (*active)[4],(*nvst)[4];

static void set_active(int *bs)
{
   int mu,x0,x1,x2,x3,ix;
   int r[4],r0,r1,r2,r3,s0,s1,s2,s3;

   r[0]=1;
   r[1]=bs[0];
   r[2]=bs[1]*r[1];
   r[3]=bs[2]*r[2];
   
   for (mu=0;mu<4;mu++)
   {
      s0=bs[mu];
      s1=bs[(mu+1)&0x3];
      s2=bs[(mu+2)&0x3];
      s3=bs[(mu+3)&0x3];

      r0=r[mu];
      r1=r[(mu+1)&0x3];
      r2=r[(mu+2)&0x3];
      r3=r[(mu+3)&0x3];      

      for (x1=0;x1<s1;x1++)
      {
         for (x2=0;x2<s2;x2++)
         {
            for (x3=0;x3<s3;x3++)
            {
               ix=r1*x1+r2*x2+r3*x3;

               if ((x1>0)&&(x1<(s1-1))&&
                   (x2>0)&&(x2<(s2-1))&&
                   (x3>0)&&(x3<(s3-1)))
               {
                  for (x0=0;x0<(s0-1);x0++)
                  {
                     active[ix][mu]=1;
                     ix+=r0;
                  }

                  active[ix][mu]=0;
               }
               else
               {
                  for (x0=0;x0<s0;x0++)
                  {
                     active[ix][mu]=0;
                     ix+=r0;
                  }
               }
            }
         }
      }
   }
}


static void set_nvst2zero(int *bs)
{
   int mu,ix,vol;

   vol=bs[0]*bs[1]*bs[2]*bs[3];

   for (ix=0;ix<vol;ix++)
   {
      for (mu=0;mu<4;mu++)
         nvst[ix][mu]=0;
   }
}


static void add_new(int *bs,int *svec)
{
   int mu,ix,iy;
   int r0,r1,r2,r3,s0,s1,s2,s3;
   int x0,x1,x2,x3,y0,y1,y2,y3;

   s0=bs[0];
   s1=bs[1];
   s2=bs[2];
   s3=bs[3];

   r0=1;
   r1=bs[0];
   r2=bs[1]*r1;
   r3=bs[2]*r2;

   for (x0=0;x0<s0;x0++)
   {
      for (x1=0;x1<s1;x1++)
      {
         for (x2=0;x2<s2;x2++)
         {
            for (x3=0;x3<s3;x3++)
            {
               ix=r0*x0+r1*x1+r2*x2+r3*x3;

               y0=x0+svec[0];
               y1=x1+svec[1];
               y2=x2+svec[2];
               y3=x3+svec[3];               

               if (y0>=s0)
                  y0-=s0;
               if (y1>=s1)
                  y1-=s1;
               if (y2>=s2)
                  y2-=s2;
               if (y3>=s3)
                  y3-=s3;

               iy=r0*y0+r1*y1+r2*y2+r3*y3;
               
               for (mu=0;mu<4;mu++)
                  nvst[iy][mu]+=active[ix][mu];
            }
         }
      }
   }
}


int main(int argc,char *argv[])
{
   int my_rank,n,ns,ix,mu;
   int bs[4],nvec;
   int vol,level,nup;
   int svec[4],rvec[4];
   int mn[4],mx[4];
   double avg[4],dev[4];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check16.log","w",stdout);
      fin=freopen("check16.in","r",stdin);
      
      printf("\n");
      printf("Check of the block shift generator\n");
      printf("----------------------------------\n\n");

      read_line("bs","%d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("level","%d\n",&level);
      read_line("nup","%d",&nup);
      fclose(fin);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&level,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nup,1,MPI_INT,0,MPI_COMM_WORLD);   

   start_ranlux(0,level);   
   geometry();
   alloc_ud();
   alloc_bgr(HMC_BLOCKS,bs);

   nvec=start_bshift(bs);
   vol=bs[0]*bs[1]*bs[2]*bs[3];

   if (my_rank==0)
   {
      printf("bs    = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nvec  = %d\n",nvec);
      printf("level = %d\n",level);
      printf("nup   = %d\n\n",nup);
      fflush(flog);

      nvst=amalloc(vol*sizeof(*nvst),3);
      active=amalloc(vol*sizeof(*active),3);
      
      error_root((nvst==NULL)||(active==NULL),1,
                 "main [check16.c]","Unable to allocate auxiliary array");

      set_active(bs);
      set_nvst2zero(bs);   
   }
   else
   {
      nvst=NULL;
      active=NULL;
   }

   rvec[0]=0;
   rvec[1]=0;
   rvec[2]=0;
   rvec[3]=0;   
   
   for (n=0;n<nup;n++)
   {
      next_bshift(svec);

      rvec[0]=(rvec[0]+svec[0])%bs[0];
      rvec[1]=(rvec[1]+svec[1])%bs[1];
      rvec[2]=(rvec[2]+svec[2])%bs[2];
      rvec[3]=(rvec[3]+svec[3])%bs[3];

      if (my_rank==0)
      {
         add_new(bs,rvec);

         if (n<(4*nvec))
         {
            printf("n = %2d, shift vector = (%2d,%2d,%2d,%2d)\n",
                   n,svec[0],svec[1],svec[2],svec[3]);
         }
      }
   }

   error_chk();
   
   if (my_rank==0)
   {
      for (mu=0;mu<4;mu++)
      {
         avg[mu]=0.0;
         dev[mu]=0.0;
         mn[mu]=nup;
         mx[mu]=0;
         
         for (ix=0;ix<vol;ix++)
         {
            n=nvst[ix][mu];
            avg[mu]+=(double)(n);
            dev[mu]+=(double)(n)*(double)(n);

            if (n<mn[mu])
               mn[mu]=n;
            if (n>mx[mu])
               mx[mu]=n;
         }

         avg[mu]/=(double)(vol);
         dev[mu]/=(double)(vol);
         dev[mu]-=(avg[mu]*avg[mu]);
      }

      printf("...\n\n");      
      printf("Minimum/Average/Maximum visiting frequency:\n");

      for (mu=0;mu<4;mu++)
      {
         printf("mu=%d,  %d/%.1f/%d, ",mu,mn[mu],avg[mu],mx[mu]);
         printf("standard deviation = %.1f\n",sqrt(dev[mu]));
      }

      printf("\n");
      printf("Perform 4*nvec shifts using mvfld:\n");
   }

   bs[0]=4;
   bs[1]=4;
   bs[2]=4;
   bs[3]=4;

   start_bshift(bs);
   
   for (n=0;n<(4*nvec);n++)
   {
      ns=mvfld();

      if (my_rank==0)
         printf("n=%2d, no of shifts=%d\n",n,ns);
   }
   
   if (my_rank==0)
   {
      printf("\n");
      fclose(flog);
   }
 
   MPI_Finalize();   
   exit(0);
}
