
/*******************************************************************************
*
* File check18.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the programs in the module chrono.c
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "block.h"
#include "hmc.h"
#include "global.h"


static double cmp_sd(int vol,spinor_dble *sd1,spinor_dble *sd2)
{
   double rd,sm,*rs1,*rs2,*rsm;

   sm=0;
   rs1=(double*)(sd1);
   rs2=(double*)(sd2);   
   rsm=rs1+24*vol;

   for (;rs1<rsm;)
   {
      rd=(*rs1)-(*rs2);
      sm+=rd*rd;

      rs1+=1;
      rs2+=1;
   }

   return sqrt(sm);
}


int main(int argc,char *argv[])
{
   int my_rank,n,k,ie;
   int bs_hmc[4],nb,isw,ib,iwd;
   int cno0,cno1,cno_blk,cno;
   double *rdold,*rdnew,rd,d1,d2,d1max,d2max;
   complex_dble z;
   spinor_dble **wsd,**sdold;
   block_t *b;
   hmc_parms_t hmc;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check18.log","w",stdout);
      fin=freopen("check18.in","r",stdin);
      
      printf("\n");
      printf("Check of the programs in the module chrono.c\n");
      printf("--------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("cno0","%d\n",&cno0);       
      read_line("cno1","%d\n",&cno1);       
      read_line("cno_blk","%d\n",&cno_blk);

      fclose(fin);
   }

   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(&cno0,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno1,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&cno_blk,1,MPI_INT,0,MPI_COMM_WORLD);
   
   start_ranlux(0,1234);
   geometry();

   hmc=set_hmc_parms(bs_hmc,-1,1,1,1,0,cno0,cno1,cno_blk,
                     1,1,1,1,1,0.5,1.0,1.0,1.0,1.0,1.0);

   if (cno0>cno1)
      cno=cno0;
   else
      cno=cno1;

   if ((2*cno)<cno_blk)
      cno=(cno_blk+1)/2;

   alloc_sd(2*cno+3);
   rdold=amalloc(4*cno*sizeof(*rdold),3);   
   sdold=amalloc(2*cno*sizeof(*sdold),3);

   error((rdold==NULL)||(sdold==NULL),1,"main [check18.c]",
         "Unable to allocate auxiliary arrays");

   rdnew=rdold+cno_blk;
   
   if (my_rank==0)
   {
      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("cno0 = %d\n",hmc.cno0);
      printf("cno1 = %d\n",hmc.cno1);
      printf("cno_blk = %d\n\n",hmc.cno_blk);
      fflush(flog);
   }

   wsd=reserve_wsd(2*cno+3);
   ie=0;
   
   for (n=0;n<cno0;n++)
   {
      if (old_chrono(0,sdold)!=n)
         ie=1;
      
      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(0,wsd[n]);
   }

   for (n=cno0;n<(cno0+3);n++)
   {
      if (old_chrono(0,sdold)!=cno0)
         ie=2;
      
      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(0,wsd[n]);
   }

   error(ie!=0,1,"main [check18.c]",
         "Wrong field count (1st global solution, ie=%d)",ie);
   
   old_chrono(0,sdold);

   for (n=0;n<cno0;n++)
   {
      if (cmp_sd(VOLUME,sdold[n],wsd[cno0+2-n])!=0.0)
         ie=1;
   }

   error(ie!=0,1,"main [check18.c]",
         "Fields are not correctly stored (1st global solution)");   

   for (n=0;n<cno1;n++)
   {
      if (old_chrono(1,sdold)!=n)
         ie=1;
      
      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(1,wsd[n]);
   }

   for (n=cno1;n<(cno1+3);n++)
   {
      if (old_chrono(1,sdold)!=cno1)
         ie=2;
      
      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(1,wsd[n]);
   }

   error(ie!=0,1,"main [check18.c]",
         "Wrong field count (2nd global solution, ie=%d)",ie);
   
   old_chrono(1,sdold);

   for (n=0;n<cno1;n++)
   {
      if (cmp_sd(VOLUME,sdold[n],wsd[cno1+2-n])!=0.0)
         ie=1;
   }

   error(ie!=0,1,"main [check18.c]",
         "Fields are not correctly stored (2nd global solution)");   
   
   reset_chrono();

   error(old_chrono(0,sdold)!=0,1,"main [check18.c]",
         "Number of old fields is not correctly reset (1st global solution)");

   error(old_chrono(1,sdold)!=0,1,"main [check18.c]",
         "Number of old fields is not correctly reset (2nd global solution)");

   for (n=0;n<(2*cno+2);)
   {
      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(0,wsd[n]);
      n+=1;

      random_sd(VOLUME,wsd[n],1.0);
      add_chrono(1,wsd[n]);
      n+=1;      
   }
      
   old_chrono(0,sdold);

   for (n=0;n<cno0;n++)
      ie+=cmp_sd(VOLUME,sdold[n],wsd[2*cno-2*n]);

   error(ie!=0,1,"main [check18.c]",
         "Fields are not correctly stored (1st global solution, mixed stores)");
   
   old_chrono(1,sdold);
   
   for (n=0;n<cno1;n++)
      ie+=cmp_sd(VOLUME,sdold[n],wsd[2*cno+1-2*n]);   

   error(ie!=0,1,"main [check18.c]",
         "Fields are not correctly stored (2nd global solution, mixed stores)");
   
   b=blk_list(HMC_BLOCKS,&nb,&isw);

   error((*b).nsd<(cno_blk+2),1,"main [check18.c]",
         "Insufficient number of double-precision spinor fields "
         "on the HMC_BLOCKS grid"); 
   
   for (ib=0;ib<nb;ib++)
   {
      for (n=0;n<cno_blk;n++)
      {
         if (old_chrono_blk(ib,rdold,sdold)!=n)
            ie=1;
         
         iwd=(wsd[n]-psd[0][0])/NSPIN;
         assign_sd2sdb(b+ib,EVEN_PTS,iwd,n+2);
         add_chrono_blk(ib,(b[ib]).sd[n+2]);
      }
   }
      
   for (ib=0;ib<nb;ib++)
   {
      for (n=cno_blk;n<(cno_blk+3);n++)
      {
         if (old_chrono_blk(ib,rdold,sdold)!=cno_blk)
            ie=2;
         
         iwd=(wsd[n]-psd[0][0])/NSPIN;
         assign_sd2sdb(b+ib,EVEN_PTS,iwd,n-cno_blk+2);
         add_chrono_blk(ib,(b[ib]).sd[n-cno_blk+2]);
      }
   }

   error(ie!=0,1,"main [check18.c]",
         "Wrong field count (block solutions, ie=%d)",ie);

   d1max=0.0;
   d2max=0.0;
   
   for (ib=0;ib<nb;ib++)
   {
      old_chrono_blk(ib,rdold,sdold);
      
      for (n=0;n<cno_blk;n++)
      {
         iwd=(wsd[cno_blk+2-n]-psd[0][0])/NSPIN;
         assign_sd2sdb(b+ib,EVEN_PTS,iwd,2+n);

         for (k=0;k<n;k++)
         {
            z=spinor_prod_dble((*b).volh,0,(b[ib]).sd[2+k],(b[ib]).sd[2+n]);
            rd=1.0/(rdnew[k]*rdnew[k]);
            z.re=-z.re*rd;
            z.im=-z.im*rd;
            mulc_spinor_add_dble((*b).volh,(b[ib]).sd[2+n],(b[ib]).sd[2+k],z);
         }

         rd=norm_square_dble((*b).volh,0,(b[ib]).sd[2+n]);
         rdnew[n]=sqrt(rd);

         d1=cmp_sd((*b).volh,(b[ib]).sd[2+n],sdold[n]);
         d2=fabs(rdnew[n]-rdold[n]);
         d1/=rdold[n];
         d2/=rdold[n];

         if (d1>d1max)
            d1max=d1;
         if (d2>d2max)
            d2max=d2;
      }
   }

   d1=d1max;
   d2=d2max;
   
   MPI_Reduce(&d1,&d1max,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Reduce(&d2,&d2max,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
   MPI_Bcast(&d1max,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   MPI_Bcast(&d2max,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     

   reset_chrono_blk();

   for (ib=0;ib<nb;ib++)
   {
      if (old_chrono_blk(ib,rdold,sdold)!=0)
         ie=1;
   }
   
   error(ie!=0,1,"main [check18.c]",
         "Number of old fields is not correctly reset (block solutions)");
   
   error_chk();

   if (my_rank==0)
   {
      printf("Global solutions: no errors detected\n\n");
      
      printf("Block solutions:\n");
      printf("Deviation of the orthogonalized solutions = %.1e\n",d1max);
      printf("Deviation of the norm of these = %.1e\n",d2max);
      printf("(should be <=1.0e-15 or so)\n\n");
      printf("No errors detected in other tests\n\n");
      fclose(flog);
   }

   MPI_Finalize();    
   exit(0);
}
