
/*******************************************************************************
*
* File check7.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Initialization of the pseudo-fermion field chi
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"
   
static complex_dble var[144],var_all[144];
static const spinor_dble sd0={{{0.0}}};


static void set_ibnd2zero(block_t *b,int k)
{
   int ifc,ix,iy,*ipp;
   spinor_dble *sk;
   bndry_t *bb;
   
   sk=(*b).sd[k];
   
   for (ifc=0;ifc<8;ifc++)
   {
      bb=(*b).bb+ifc;
      ipp=(*bb).ipp;

      for (ix=0;ix<(*bb).vol;ix++)
      {
         iy=ipp[ix];
         sk[iy]=sd0;
      }  
   }
}


static int is_on_edge(block_t *b,int ix)
{
   int mu,it;

   it=0;
   
   for (mu=0;mu<4;mu++)
   {
      if ((*b).iup[ix][mu]>=(*b).vol)
         it+=1;
      if ((*b).idn[ix][mu]>=(*b).vol)
         it+=1;
   }

   if (it<=1)
      return 0;
   else
      return 1;
}


static double check_bnd(block_t *b,int k)
{
   int ifc,ix,iy,i,*ipp;
   double d,dmax,*s;
   weyl_dble *sk;
   bndry_t *bb;

   dmax=0.0;
   
   for (ifc=0;ifc<8;ifc++)
   {
      bb=(*b).bb+ifc;
      ipp=(*bb).ipp;
      sk=(*bb).sd[k];

      for (ix=0;ix<(*bb).vol;ix++)
      {
         iy=ipp[ix];

         if (is_on_edge(b,iy)==0)
         {
            s=(double*)(sk+ix);

            for (i=0;i<12;i++)
            {
               d=fabs(s[i]);
               if (d>dmax)
                  dmax=d;
            }
         }
      }
   }

   return dmax;
}


int main(int argc,char *argv[])
{
   int my_rank,nb,nbh,isw;
   int n,i,j,k,ntst,bs[4];
   double d,dmax,dmax_all,r;
   complex_dble *z;
   spinor_dble *s,*sm;
   block_t *b;
   hmc_t *hflds;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check7.log","w",stdout);
      fin=freopen("check7.in","r",stdin);
      
      printf("\n");
      printf("Initialization of the pseudo-fermion field chi\n");
      printf("----------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      fclose(fin);

      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);       
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);

   start_ranlux(0,1234);
   geometry();
   alloc_ud();
   alloc_sd(2);
   alloc_bgr(HMC_BLOCKS,bs);
   
   random_ud();
   assign_ud2udbgr(HMC_BLOCKS);
   blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();
   nbh=nb/2;
   
   for (i=0;i<12;i++)
   {
      for (j=i;j<12;j++)
      {
         var[12*i+j].re=0.0;
         var[12*i+j].im=0.0;
      }
   }

   ntst=100000/(nbh*hflds[0].nbnd);

   if (ntst<1)
      ntst=1;
   
   for (k=0;k<ntst;k++)
   {
      random_chi();      

      for (n=0;n<nbh;n++)
      {
         s=hflds[n+isw*nbh].chi;
         sm=s+hflds[0].nbnd;

         for (;s<sm;s++)
         {
            z=(complex_dble*)(s);

            for (i=0;i<12;i++)
            {
               for (j=i;j<12;j++)
               {
                  var[12*i+j].re+=z[i].re*z[j].re+z[i].im*z[j].im;
                  var[12*i+j].im+=z[i].im*z[j].re-z[i].re*z[j].im;
               }
            }            
         }
      }
   }

   n=hflds[0].nbnd-
      (bs[0]-2)*(bs[1]-2)*(bs[2]-2)-
      (bs[1]-2)*(bs[2]-2)*(bs[3]-2)-
      (bs[2]-2)*(bs[3]-2)*(bs[0]-2)-
      (bs[3]-2)*(bs[0]-2)*(bs[1]-2);
   
   r=1.0/((double)(NPROC*ntst*n*nbh));
   
   for (i=0;i<12;i++)
   {
      for (j=i;j<12;j++)
      {
         var[12*i+j].re*=r;
         var[12*i+j].im*=r;
      }
   }

   MPI_Reduce((double*)(var),(double*)(var_all),288,MPI_DOUBLE,
              MPI_SUM,0,MPI_COMM_WORLD);
   
   if (my_rank==0)
   {
      printf("Statistical test of the distribution of chi:\n");
      
      dmax=0.0;
      
      for (i=0;i<12;i++)
      {
         for (j=i;j<12;j++)
         {
            if (i==j)
            {
               printf("<chi_%d*chi_%d^*> \t= % .4e ",
                      i,j,var_all[12*i+j].re);
               printf("(expect 1.0 within errors)\n");

               d=fabs(var_all[12*i+j].im);
               if (d>dmax)
                  dmax=d;
            }
            else
            {
               d=fabs(var_all[12*i+j].re);
               if (d>dmax)
                  dmax=d;
               d=fabs(var_all[12*i+j].im);
               if (d>dmax)
                  dmax=d;
            }
         }
      }

      printf("\n");
      printf("max|<chi_i*chi_j^*>| = %.1e ",dmax);
      printf("(expect 0.0 within errors)\n\n");
   }

   random_chi();   
   assign_chi2sd(0);
   assign_sd2sd(VOLUME,psd[0][0],psd[1][0]);
   sub_assign_PdOs(1,0);
   dmax=norm_square_dble(VOLUME,0,psd[1][0])/
      norm_square_dble(VOLUME,0,psd[0][0]);
   dmax=sqrt(dmax);

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);   

   if (my_rank==0)
   {
      printf("Comparison of chi with PdOmega^star*chi:\n");
      printf("Maximal relative deviation = %.1e\n\n",
             dmax_all);
   }   

   random_sd(VOLUME,psd[0][0],1.0);
   random_chi();   
   assign_chi2sd(0);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   dmax=0.0;
   
   for (n=0;n<nb;n++)
   {
      assign_sd2sdb(b,ALL_PTS,0,0);

      if (((isw==0)&&(n<nbh))||((isw==1)&&(n>=nbh)))
         set_ibnd2zero(b,0);

      d=norm_square_dble((*b).vol,0,(*b).sd[0]);
      if (d>dmax)
         dmax=d;
      b+=1;
   }

   dmax=sqrt(dmax);
   
   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD); 

   if (my_rank==0)
   {
      printf("Check that chi is supported on dOmega^star:\n");
      printf("Maximal size away from dOmega^star = %.1e\n\n",
             dmax_all);
   }

   random_chi();   
   assign_chi2sd(0);
   b=blk_list(HMC_BLOCKS,&nb,&isw);
   dmax=0.0;
   
   for (n=0;n<nb;n++)
   {
      assign_sd2sdb(b,ALL_PTS,0,0);
      mulg5_dble((*b).vol,(*b).sd[0]);

      Qbnds_dble(b,0,0);
      d=check_bnd(b,0);
      if (d>dmax)
         dmax=d;
      b+=1;
   }

   error_chk();
   
   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD); 

   if (my_rank==0)
   {
      printf("Check that DdOmega^star*gamma5*chi vanishes where it should:\n");
      printf("Maximal size = %.1e\n\n",dmax_all);
      fclose(flog);   
   }

   MPI_Finalize();   
   exit(0);
}
