
/*******************************************************************************
*
* File time4.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the valg_dble routines
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "dfl.h"
#include "global.h"

static int nv;
static complex_dble *vmat,*wmat;
static complex_dble *ppk[10],**pvd;


static double wt_vprod_dble(int nflds,int icom)
{
   int my_rank,nmax,n,i,ib;
   complex_dble z;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   

   for (i=0;i<nflds;i++)
      random_vd(nv,pvd[i],1.0);

   nmax=1;
   
   for (ib=0;ib<1;nmax*=2)
   {   
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i+=2)
            z=vprod_dble(nv,icom,pvd[i],pvd[i+1]);
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)((nmax*nflds)/2);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   return wtav;
}


static double wt_vnorm_square_dble(int nflds,int icom)
{
   int my_rank,nmax,n,i,ib;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i++)
            (void)(vnorm_square_dble(nv,icom,pvd[i]));
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(nmax*nflds);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);      
   
   return wtav;
}


static double wt_vnormalize_dble(int nflds,int icom)
{
   int my_rank,nmax,n,i,ib;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i++)
            (void)(vnormalize_dble(nv,icom,pvd[i]));
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(nmax*nflds);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);      
   
   return wtav;
}


static double wt_mulc_vadd_dble(int nflds)
{
   int my_rank,nmax,n,i,ib;
   complex_dble z;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   
   z.re=0.123f;
   z.im=0.456f;
   nmax=1;
   
   for (ib=0;ib<1;nmax*=2)
   {   
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i+=2)
            mulc_vadd_dble(nv,pvd[i],pvd[i+1],z);

         z.re-=z.re;
         z.im-=z.im;
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)((nmax*nflds)/2);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   return wtav;
}


static double wt_vproject_dble(int nflds,int icom)
{
   int my_rank,nmax,n,i,ib;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   

   for (i=0;i<nflds;i++)
   {
      random_vd(nv,pvd[i],1.0);
      (void)(vnormalize_dble(nv,icom,pvd[i]));
   }

   nmax=1;
   
   for (ib=0;ib<1;nmax*=2)
   {   
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         for (i=0;i<nflds;i+=2)
            vproject_dble(nv,icom,pvd[i],pvd[i+1]);
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)((nmax*nflds)/2);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   
   return wtav;
}


static void gram_schmidt(int n,complex_dble **vd)
{
   int i,j,k;

   for (i=0;i<n;i++)
   {
      for (k=0;k<2;k++)
      {
         for (j=0;j<i;j++)
            vproject_dble(nv,0,vd[i],vd[j]);
      }

      (void)(vnormalize_dble(nv,0,vd[i]));
   }
}


static void random_unitary(int n)
{
   int i,j,k;

   vmat=amalloc(2*n*n*sizeof(*vmat),4);
   wmat=vmat+n*n;
   error(vmat==NULL,1,"random_unitary [time4.c]",
         "Unable to allocate auxiliary arrays");
   
   gauss_dble((double*)(vmat),2*n*n);

   for (i=0;i<n;i++)
   {
      for (k=0;k<2;k++)
      {
         for (j=0;j<i;j++)
            vproject_dble(n,0,vmat+i*n,vmat+j*n);
      }

      (void)(vnormalize_dble(n,0,vmat+i*n));
   }   

   for (i=0;i<n;i++)
   {
      for (j=0;j<n;j++)
      {
         wmat[i*n+j].re=vmat[j*n+i].re;
         wmat[i*n+j].im=-vmat[j*n+i].im;         
      }
   }
}


static double wt_vrotate_dble(void)
{
   int my_rank,nmax,n,i,ib;
   double wt1,wt2,wdt,wtav;

   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);   

   for (i=0;i<5;i++)
   {
      random_vd(nv,pvd[i],1.0);
      ppk[i]=pvd[i];
   }

   gram_schmidt(5,ppk);
   for (i=0;i<5;i++)
      assign_vd2vd(nv,pvd[i],pvd[i+5]);
   random_unitary(5);
   nmax=1;

   for (ib=0;ib<1;nmax*=2)
   {
      MPI_Barrier(MPI_COMM_WORLD);
      wt1=MPI_Wtime();

      for (n=0;n<nmax;n++)
      {
         if ((n&0xf)==0x0)
         {
            for (i=0;i<5;i++)
               assign_vd2vd(nv,pvd[i+5],pvd[i]);
         }
      
         vrotate_dble(nv,5,ppk,vmat);
         vrotate_dble(nv,5,ppk,wmat);
      }
      
      wt2=MPI_Wtime();
      wdt=wt2-wt1;

      MPI_Reduce(&wdt,&wtav,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);      

      if (my_rank==0)
      {
         wtav/=(double)(NPROC);

         if (wtav>2.0)
            ib=1;

         wtav/=(double)(2*nmax);
      }

      MPI_Bcast(&ib,1,MPI_INT,0,MPI_COMM_WORLD);
   }

   MPI_Bcast(&wtav,1,MPI_DOUBLE,0,MPI_COMM_WORLD);      
   
   return wtav;
}


int main(int argc,char *argv[])
{
   int my_rank,icom,nflds;
   int bs[4],nm,n,nvec;
   double wdt;
   FILE *flog=NULL,*fin=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("time4.log","w",stdout);
      fin=freopen("time4.in","r",stdin);
      
      printf("\n");
      printf("Timing of the valg routines\n");
      printf("---------------------------\n\n");   
      
      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bs","%d %d %d %d",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nm","%d",&nm);
      fclose(fin);

      printf("bs = %d %d %d %d\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nm = %d\n\n",nm);
   }

   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nm,1,MPI_INT,0,MPI_COMM_WORLD);   

   icom=1;   
   start_ranlux(0,12345);
   geometry();
   set_dfl_parms(bs,nm,4,1,1,1,1,1,1,0.1,1.0e-3,1.0e-6,1.0e-12,0,0);
   (void)(vdflds_base(&n,&nv,&nvec));   
   
   if (my_rank==0)
   {   
      if (NPROC>1)
         printf("There are %d MPI processes\n",NPROC);
      else
         printf("There is 1 MPI process\n");
      
      if ((int)(nv*sizeof(complex_dble))<(64*1024))
         printf("The local size of a vector field is %d KB\n",
                (int)((nv*sizeof(complex_dble))/(1024)));
      else
         printf("The local size of a vector field is %d MB\n",
                (int)((nv*sizeof(complex_dble))/(1024*1024)));

#if (defined SSE3)
      printf("Using inline assembly SSE3 instructions\n");
#elif (defined SSE2)
      printf("Using inline assembly SSE2 instructions\n");      
#elif (defined SSE)
      printf("Using inline assembly SSE instructions\n");
#endif

#if (defined SSE)
#if (defined P3)
      printf("Assuming SSE prefetch instructions fetch 32 bytes\n");
#elif (defined PM)
      printf("Assuming SSE prefetch instructions fetch 64 bytes\n");
#elif (defined P4)
      printf("Assuming SSE prefetch instructions fetch 128 bytes\n");
#else
      printf("SSE prefetch instructions are not used\n");
#endif
#endif
      
      printf("\n");
   }

   nflds=(int)((32*1024*1024)/(int)(nv*sizeof(double)))+1;
   if ((nflds%2)==1)
      nflds+=1;
   if (nflds<10)
      nflds=10;

   set_dfl_parms(bs,nm,4,1,1,1,1,1,1,0.1,1.0e-3,1.0e-6,1.0e-12,0,nflds);
   pvd=reserve_wvd(nflds);
   
   wdt=1.0e6*wt_vprod_dble(nflds,icom)/(double)(nv);   

   if (my_rank==0)
   {   
      printf("Function vprod_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(8.0/wdt),4*(int)(sizeof(complex_dble)));
   }

   wdt=1.0e6*wt_vnorm_square_dble(nflds,icom)/(double)(nv);   

   if (my_rank==0)
   {   
      printf("Function vnorm_square_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(4.0/wdt),4*(int)(sizeof(complex_dble)));
   }

   wdt=1.0e6*wt_vnormalize_dble(nflds,icom)/(double)(nv);   
   
   if (my_rank==0)
   {   
      printf("Function vnormalize_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(6.0/wdt),4*(int)(sizeof(complex_dble)));
   }
   
   wdt=1.0e6*wt_mulc_vadd_dble(nflds)/(double)(nv);   
   
   if (my_rank==0)
   {   
      printf("Function mulc_vadd_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(8.0/wdt),4*(int)(sizeof(complex_dble)));
   }

   wdt=1.0e6*wt_vproject_dble(nflds,icom)/(double)(nv);   
   
   if (my_rank==0)
   {   
      printf("Function vproject_dble:\n");
      printf("Time per block: %4.3f micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(16.0/wdt),4*(int)(sizeof(complex_dble)));
   }
   
   wdt=1.0e6*wt_vrotate_dble()/(double)(25*nv);
   error_chk();   

   if (my_rank==0)
   {   
      printf("Function vrotate_dble (n=5 fields):\n");
      printf("Time per block: %4.3f*n^2 micro sec\n",(double)(nm)*wdt);
      printf("%d Mflops [%d bit arithmetic]\n\n",
             (int)(7.6/wdt),4*(int)(sizeof(complex_dble)));
      fclose(flog);
   }

   MPI_Finalize();   
   exit(0);
}
