
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005, 2007, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Importing a configuration previously exported by check3, possibly with
* periodic extension
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "misc.h"
#include "start.h"
#include "update.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,nsize,ir;
   stdint_t l[4];
   float beta;
   double plaq0,plaq1,plaq2;
   char cnfg_dir[NAME_SIZE],name[NAME_SIZE];
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      flog=freopen("check4.log","w",stdout);
      fin=freopen("check3.in","r",stdin);
      
      printf("\n");
      printf("Importing a configuration previously exported by check3\n");
      printf("-------------------------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n",L0,L1,L2,L3);

      read_line("cnfg_dir","%s\n",cnfg_dir);
      fclose(fin);

      fflush(flog);
   }

   MPI_Bcast(cnfg_dir,NAME_SIZE,MPI_CHAR,0,MPI_COMM_WORLD);
   
   beta=5.7f;
   start_ranlux(0,123456);
   geometry();
   alloc_u();
   alloc_ud();
   random_u();

   update(beta,8,0,8);
   assign_u2ud();
   plaq0=plaq_sum_dble()/(double)(6*NPROC*VOLUME);

   check_dir_root(cnfg_dir);   
   nsize=name_size("%s/testcnfg",cnfg_dir);
   error_root(nsize>=NAME_SIZE,1,"main [check4.c]","cnfg_dir name is too long");
   sprintf(name,"%s/testcnfg",cnfg_dir);

   if (my_rank==0)
   {
      fin=fopen(name,"rb");
      error_root(fin==NULL,1,"main [check4.c]","Unable to open input file");

      ir=fread(l,sizeof(stdint_t),4,fin);
      ir+=fread(&plaq1,sizeof(double),1,fin);
      error_root(ir!=5,1,"main [check4.c]","Incorrect read count");
      fclose(fin);

      if (endianness()==BIG_ENDIAN)
      {
         bswap_int(4,l);
         bswap_double(1,&plaq1);
      }
      
      printf("Random configuration, average plaquette = %.12f\n\n",plaq0);
      printf("Now read configuration from file %s\n",name);
      printf("%dx%dx%dx%d lattice\n",
             (int)(l[0]),(int)(l[1]),(int)(l[2]),(int)(l[3]));
      printf("Average plaquette = %.12f\n",plaq1);
   }

   import_cnfg(name);
   plaq2=plaq_sum_dble()/(double)(6*NPROC*VOLUME);
   error_chk();

   if (my_rank==0)
   {
      printf("Check             = %.12f\n\n",plaq2);
      remove(name);
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
