
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program jacobi1_dble
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "eva.h"


int main(void)
{
   int n,i,j,k;
   double *a,*d,*v,*as;
   double *av,*vav,p,r1,r2,r3;

   printf("\n");
   printf("Diagonalization of a symmetric matrix with random elements\n");
   printf("----------------------------------------------------------\n\n");
   printf("(Double-precision program)\n\n");
   
   rlxd_init(1,1234);   

   for (;;)
   {
      printf("Select rank n of matrix: ");
      scanf("%d",&n);

      a=amalloc((5*n*n+n)*sizeof(double),3);
      error(a==NULL,1,"main [check3.c]","Unable to allocate auxiliary arrays");

      d=a+n*n;
      v=d+n;
      as=v+n*n;
      av=as+n*n;
      vav=av+n*n;

      gauss_dble(a,n*n);

      for (i=0;i<n;i++)
      {
         as[i*n+i]=a[i*n+i];
         
         for (j=(i+1);j<n;j++)
         {
            a[i*n+j]=a[j*n+i];

            as[i*n+j]=a[i*n+j];
            as[j*n+i]=a[j*n+i];
         }
      }
      
      jacobi1_dble(n,a,d,v);

      j=0;
      
      for (i=0;i<(n*n);i++)
         j+=(a[i]!=as[i]);

      error(j!=0,1,"main [check3.c]","Input matrix is not preserved"); 
      
      printf("\n");
      printf("The first 10 eigenvalues are:\n\n");

      for (i=0;(i<n)&&(i<10);i++)
         printf("%d   %#14.6e\n",i,d[i]);
      
      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p=0.0;

            for (k=0;k<n;k++)
               p+=a[n*i+k]*v[n*k+j];

            av[n*i+j]=p;
         }
      }

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p=0.0;

            for (k=0;k<n;k++)
               p+=v[n*k+i]*av[n*k+j];

            vav[n*i+j]=p;
         }
      }

      r1=0.0;
      r2=0.0;
      r3=0.0;

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            if (i!=j)
            {
               p=fabs(vav[n*i+j]);
               if (p>r1)
                  r1=p;
            }
            else
            {
               p=fabs(d[i]-vav[n*i+j]);
               if (p>r2)
                  r2=p;
               p=fabs(d[i]);
               if (p>r3)
                  r3=p;
            }
         }
      }

      r1/=r3;
      r2/=r3;

      printf("\n");
      printf("Verification of the diagonalization:\n");
      printf("Max{(off-diagonal elements)/(norm of the matrix)} = %.1e\n",r1);
      printf("Max{(deviation on diagonal)/(norm of the matrix)} = %.1e\n\n",r2);

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p=0.0;
            
            for (k=0;k<n;k++)
               p+=v[n*k+i]*v[n*k+j];

            vav[n*i+j]=p;
         }
      }

      r1=0.0;
      r2=0.0;

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            if (i!=j)
            {
               p=fabs(vav[n*i+j]);
               if (p>r1)
                  r1=p;
            }
            else
            {
               p=fabs(1.0-vav[n*i+j]);
               if (p>r2)
                  r2=p;
            }
         }
      }

      printf("Verification of the orthonormality of the eigenvectors:\n");
      printf("Max{((v_i,v_j) for i different from j)} = %.1e\n",r1);
      printf("Max{(deviation of (v_i,v_i) from 1)}    = %.1e\n",r2);
      printf("\n");

      afree(a);
   }
   
   exit(0);
}
