
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the program jacobi2_dble
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "eva.h"


int main(void)
{
   int n,i,j,k;
   double *d;
   double r1,r2,r3,q;
   complex_dble *a,*v,*as;
   complex_dble p,*av,*vav;

   printf("\n");
   printf("Diagonalization of a hermitean matrix with random elements\n");
   printf("----------------------------------------------------------\n\n");
   printf("(Double-precision program)\n\n");
   
   rlxd_init(1,4321);   
   
   for (;;)
   {
      printf("Select rank n of matrix: ");
      scanf("%d",&n);

      a=amalloc(5*n*n*sizeof(complex_dble),3);
      d=amalloc(n*sizeof(double),3);
      
      error((a==NULL)||(d==NULL),1,"main [check4.c]",
            "Unable to allocate auxiliary arrays");

      v=a+n*n;
      as=v+n*n;
      av=as+n*n;
      vav=av+n*n;
      
      gauss_dble((double*)(a),2*n*n);

      for (i=0;i<n;i++)
      {
         a[i*n+i].im=0.0;
         as[i*n+i]=a[i*n+i];
         
         for (j=(i+1);j<n;j++)
         {
            a[i*n+j].re= a[j*n+i].re;
            a[i*n+j].im=-a[j*n+i].im;
            
            as[i*n+j]=a[i*n+j];
            as[j*n+i]=a[j*n+i];
         }
      }

      jacobi2_dble(n,a,d,v);

      j=0;
      
      for (i=0;i<(n*n);i++)
         j+=((a[i].re!=as[i].re)||(a[i].im!=as[i].im));

      error(j!=0,1,"main [check4.c]","Input matrix is not preserved");       
      
      printf("\n");
      printf("The first 10 eigenvalues are:\n\n");

      for (i=0;(i<n)&&(i<10);i++)
         printf("%d   %#14.6e\n",i,d[i]);

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p.re=0.0;
            p.im=0.0;
            
            for (k=0;k<n;k++)
            {
               p.re+=a[n*i+k].re*v[n*k+j].re-a[n*i+k].im*v[n*k+j].im;
               p.im+=a[n*i+k].im*v[n*k+j].re+a[n*i+k].re*v[n*k+j].im;
            }
            
            av[n*i+j]=p;
         }
      }

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p.re=0.0;
            p.im=0.0;
            
            for (k=0;k<n;k++)
            {
               p.re+=v[n*k+i].re*av[n*k+j].re+v[n*k+i].im*av[n*k+j].im;
               p.im+=v[n*k+i].re*av[n*k+j].im-v[n*k+i].im*av[n*k+j].re;
            }
            
            vav[n*i+j]=p;
         }
      }

      r1=0.0;
      r2=0.0;
      r3=0.0;

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            if (i!=j)
            {
               q=fabs(vav[n*i+j].re)+fabs(vav[n*i+j].im);
               if (q>r1)
                  r1=q;
            }
            else
            {
               q=fabs(d[i]-vav[n*i+j].re)+fabs(vav[n*i+j].im);
               if (q>r2)
                  r2=q;
               q=fabs(d[i]);
               if (q>r3)
                  r3=q;
            }
         }
      }

      r1/=r3;
      r2/=r3;

      printf("\n");
      printf("Verification of the diagonalization:\n");
      printf("Max{(off-diagonal elements)/(norm of the matrix)} = %.1e\n",r1);
      printf("Max{(deviation on diagonal)/(norm of the matrix)} = %.1e\n\n",r2);

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            p.re=0.0;
            p.im=0.0;

            for (k=0;k<n;k++)
            {
               p.re+=v[n*k+i].re*v[n*k+j].re+v[n*k+i].im*v[n*k+j].im;
               p.im+=v[n*k+i].re*v[n*k+j].im-v[n*k+i].im*v[n*k+j].re;
            }

            vav[n*i+j]=p;
         }
      }

      r1=0.0;
      r2=0.0;

      for (i=0;i<n;i++)
      {
         for (j=0;j<n;j++)
         {
            if (i!=j)
            {
               q=fabs(vav[n*i+j].re)+fabs(vav[n*i+j].im);
               if (q>r1)
                  r1=q;
            }
            else
            {
               q=fabs(1.0-vav[n*i+j].re)+fabs(vav[n*i+j].im);
               if (q>r2)
                  r2=q;
            }
         }
      }

      printf("Verification of the orthonormality of the eigenvectors:\n");
      printf("Max{((v_i,v_j) for i different from j)} = %.1e\n",r1);
      printf("Max{(deviation of (v_i,v_i) from 1)}    = %.1e\n",r2);
      printf("\n");

      afree(a);
      afree(d);
   }
   
   exit(0);
}
