
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of det2u3alg and vec2u3alg
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "hmc.h"

#define N 10000000


int main(void)
{
   int count;
   double t1,t2,dt;
   su3_dble *u,*v;
   spinor_dble *r,*s;
   pauli_dble *m;
   u3_alg_dble *X;
   
   printf("\n");
   printf("Timing of det2u3alg and vec2u3alg\n");
   printf("---------------------------------\n\n");

   u=amalloc(2*sizeof(su3_dble),4);
   r=amalloc(2*sizeof(spinor_dble),4);
   m=amalloc(2*sizeof(pauli_dble),4);
   X=amalloc(6*sizeof(u3_alg_dble),3);   
   error((u==NULL)||(r==NULL)||(m==NULL)||(X==NULL),1,
         "main [time1.c]","Unable to allocate auxiliary arrays");

   s=r+1;
   v=u+1;
   
   rlxd_init(1,23456);
   random_su3_dble(u);
   random_su3_dble(v);
   ranlxd((double*)(r),48);
   ranlxd(m[0].u,36);
   ranlxd(m[1].u,36);
   ranlxd((double*)(X),54);
   printf("The times per application are:\n");
   
   t1=(double)clock();
   for (count=0;count<(2*N);count++)
      det2u3alg(m,X);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(2*N);

   printf("det2u3alg:     %4.3f micro sec (%d Mflops)\n",dt,(int)(108.0/dt));

   t1=(double)clock();
   for (count=0;count<N;count++)
      vec2u3alg(r,s,X);
   t2=(double)clock();

   dt=(t2-t1)/(double)(CLOCKS_PER_SEC);
   dt*=1.0e6/(double)(N);

   printf("vec2u3alg:     %4.3f micro sec (%d Mflops)\n\n",dt,(int)(612.0/dt));
   exit(0);
}
