
/*******************************************************************************
*
* File cheby1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Chebyshev approximation of ln(1+x)=x*sum_n c_n*T_n(x), -0.25<=x<=0.5
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "extras.h"


static double f(double x)
{
   double y;
   
   if (fabs(x)<=1.0e-3)
   {
      y=1.0/5.0;
      y=x*y-1.0/4.0;
      y=x*y+1.0/3.0;
      y=x*y-1.0/2.0;
      y=x*y+1.0;
   }
   else
      y=log(1.0+x)/x;

   return(y);
}


int main(void)
{
   int nmax=64,n,i;
   double a,b,eps,c[64];

   printf("\n");
   printf("Chebyshev approximation of ln(1+x)/x\n");
   printf("------------------------------------\n");
   printf("\n");

   a=-0.25;
   b=0.5;
   eps=3.0e-8;

   printf("Range %.3f<=x<=%.3f, absolute accuracy %.1e\n\n",a,b,eps);
   
   n=cheby_fit(a,b,f,nmax,eps,c);

   for (i=0;i<=n;i++)
      printf("   c[%d]=% 10.9ff;\n",i,c[i]);

   printf("\n");   
   exit(0);
}
