
/*******************************************************************************
*
* File check4.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Accuracy of the functions sse_ln1px and sse_sincos
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "random.h"
#include "sse.h"
#include "misc.h"

#if (defined SSE)

#define N 4000000


static double ln1px(double x)
{
   double y;
   
   if (fabs(x)<=1.0e-3)
   {
      y=1.0/5.0;
      y=x*y-1.0/4.0;
      y=x*y+1.0/3.0;
      y=x*y-1.0/2.0;
      y=x*y+1.0;
      y*=x;
   }
   else
      y=log(1.0+x);

   return(y);
}


static double sinpix(double x)
{
   double w,y;

   w=x*4.0*atan(1.0);
   
   if (fabs(w)<=1.0e-3)
   {
      y=1.0/120.0;
      y=w*w*y-1.0/6.0;
      y=w*w*y+1.0;
      y*=w;
   }
   else
      y=sin(w);

   return(y);
}


int main(void)
{
   int i;
   double d,dmax;
   static float r[4*N] __attribute__ ((aligned (16)));
   static sse_float x,y,z,u,v;
   
   printf("\n");
   printf("SSE coding of ln(1+x), sin(pi*x) and cos(pi*x)\n");
   printf("----------------------------------------------\n\n");
   printf("Number of tested values: %d\n\n",4*N);   
   
   ranlxs(r,4*N);
   dmax=0.0;

   for (i=0;i<4*N;i+=4)
   {
      x.c1=-r[i];
      x.c2=r[i+1];
      x.c3=1.0f/(1.0f-r[i+2]);
      x.c4=1.0e-6f*(2.0f*r[i+3]-1.0f);

      if ((x.c1==0.0f)||(x.c1<=-1.0f))
         x.c1=1.0f;
      if ((x.c2==0.0f)||(x.c2<=-1.0f))
         x.c2=1.0f;
      if ((x.c3==0.0f)||(x.c3<=-1.0f))
         x.c3=1.0f;
      if ((x.c4==0.0f)||(x.c4<=-1.0f))
         x.c4=1.0f;      
      
      sse_ln1px(&x,&y);

      d=ln1px((double)(x.c1))/(double)(y.c1)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=ln1px((double)(x.c2))/(double)(y.c2)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=ln1px((double)(x.c3))/(double)(y.c3)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=ln1px((double)(x.c4))/(double)(y.c4)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
   }

   printf("Maximal relative error of ln(1+x) (x!=0.0): ");
   printf("%4.2e\n",dmax);

   x.c1=0.0f;
   x.c2=(float)(ldexp(1.0,-23));
   x.c3=-1.0f+(float)(ldexp(1.0,-24));
   x.c4=(float)(ldexp(1.0,126));

   sse_ln1px(&x,&y);

   printf("ln(1.0) = %4.2e\n",y.c1);
   d=ln1px((double)(x.c2))/(double)(y.c2)-1.0;
   printf("Relative error at x = 2^(-23): %4.2e\n",fabs(d));
   d=ln1px((double)(x.c3))/(double)(y.c3)-1.0;
   printf("Relative error at x = -1+2^(-24): %4.2e\n",fabs(d));
   d=ln1px((double)(x.c4))/(double)(y.c4)-1.0;
   printf("Relative error at x = 2^126: %4.2e\n\n",fabs(d));   
   
   dmax=0.0;

   for (i=0;i<4*N;i+=4)
   {
      x.c1=r[i]-0.5f;
      x.c2=r[i+1]-0.5f;
      x.c3=r[i+2]-0.5f;
      x.c4=r[i+3]-0.5f;      

      if (x.c1==0.0f)
         x.c1=0.5f;
      if (x.c2==0.0f)
         x.c2=0.5f;
      if (x.c3==0.0f)
         x.c3=0.5f;
      if (x.c4==0.0f)
         x.c4=0.5f;      
      
      sse_sincos(&x,&y,&z);

      d=sinpix((double)(x.c1))/(double)(y.c1)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=sinpix((double)(x.c2))/(double)(y.c2)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=sinpix((double)(x.c3))/(double)(y.c3)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      d=sinpix((double)(x.c4))/(double)(y.c4)-1.0;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
   }

   printf("Maximal relative error of sin(pi*x) (|x|<=0.5, x!=0.0): ");
   printf("%4.2e\n",dmax);

   dmax=0.0;

   for (i=0;i<4*N;i+=4)
   {
      x.c1=2.0f*r[i]-1.0f;
      x.c2=2.0f*r[i+1]-1.0f;
      x.c3=2.0f*r[i+2]-1.0f;
      x.c4=2.0f*r[i+3]-1.0f;      

      u.c1=1.0f-fabs(x.c1);
      u.c2=1.0f-fabs(x.c2);
      u.c3=1.0f-fabs(x.c3);
      u.c4=1.0f-fabs(x.c4);      
      
      sse_sincos(&x,&y,&z);
      sse_sincos(&u,&v,&z);

      if (x.c1<0.0f)
         d=v.c1+y.c1;
      else
         d=v.c1-y.c1;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      
      if (x.c2<0.0f)
         d=v.c2+y.c2;
      else
         d=v.c2-y.c2;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      
      if (x.c3<0.0f)
         d=v.c3+y.c3;
      else
         d=v.c3-y.c3;
      d=fabs(d);
      if (d>dmax)
         dmax=d;

      if (x.c4<0.0f)
         d=v.c4+y.c4;
      else
         d=v.c4-y.c4;
      d=fabs(d);
      if (d>dmax)
         dmax=d;        
   }

   printf("Maximal absolute difference sin(pi*x)-sign(x)*sin(pi*(1-|x|): ");
   printf("%4.2e\n",dmax);

   dmax=0.0;

   for (i=0;i<4*N;i+=4)
   {
      x.c1=2.0f*r[i]-1.0f;
      x.c2=2.0f*r[i+1]-1.0f;
      x.c3=2.0f*r[i+2]-1.0f;
      x.c4=2.0f*r[i+3]-1.0f;      

      u.c1=0.5f-fabs(x.c1);
      u.c2=0.5f-fabs(x.c2);
      u.c3=0.5f-fabs(x.c3);
      u.c4=0.5f-fabs(x.c4);      
      
      sse_sincos(&x,&z,&y);
      sse_sincos(&u,&v,&z);

      d=v.c1-y.c1;
      d=fabs(d);
      if (d>dmax)
         dmax=d;

      d=v.c2-y.c2;
      d=fabs(d);
      if (d>dmax)
         dmax=d;
      
      d=v.c3-y.c3;
      d=fabs(d);
      if (d>dmax)
         dmax=d;

      d=v.c4-y.c4;
      d=fabs(d);
      if (d>dmax)
         dmax=d;        
   }

   printf("Maximal absolute difference cos(pi*x)-sin(pi*(0.5-|x|): ");
   printf("%4.2e\n",dmax);

   x.c1=0.0f;
   x.c2=0.5f;
   x.c3=1.0f;
   x.c4=1.0f;

   sse_sincos(&x,&y,&z);

   printf("sin(pi*0.0) = %.8e, cos(pi*0.0) = %.8e \n",y.c1,z.c1);
   printf("sin(pi*0.5) = %.8e, cos(pi*0.5) = %.8e \n",y.c2,z.c2);
   printf("sin(pi*1.0) = %.8e, cos(pi*1.0) = %.8e \n\n",y.c3,z.c3);
   
   exit(0);
}

#else

int main(void)
{
   printf("\n");
   printf("SSE coding of ln(1+x), sin(pi*x) and cos(pi*x)\n");
   printf("----------------------------------------------\n\n");

   printf("SSE must be enabled for this test\n\n");
   exit(0);
}

#endif
