
/*******************************************************************************
*
* File time1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the functions sse_ln1px and sse_sincos
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "sse.h"
#include "misc.h"

#if (defined SSE)

#define NLOOP 10000000


int main(void)
{
   int n;
   float t1,t2,dt;
   static sse_float x,y,z;   
   
   printf("\n");
   printf("Timing of the SSE code for ln(1+x)\n");
   printf("----------------------------------\n\n");

   x.c1=-0.78654f;
   x.c2=0.0014536f;
   x.c3=8.879e7f;
   x.c4=-1.0023e-3f;
   
   t1=(float)clock();
   for (n=0;n<NLOOP;n++)
      sse_ln1px(&x,&y);
   t2=(float)clock();

   dt=(t2-t1)/(float)(CLOCKS_PER_SEC);
   dt*=1.0e6f/(float)(4*NLOOP);

   printf("Average time per value of x: %4.3f micro sec\n\n\n",dt);

   printf("Timing of the SSE code for sin(pi*x) and cos(pi*x)\n");
   printf("--------------------------------------------------\n\n");
   
   x.c1=-0.78654f;
   x.c2=0.0014536f;
   x.c3=0.555f;
   x.c4=-0.999f;
   
   t1=(float)clock();
   for (n=0;n<NLOOP;n++)
      sse_sincos(&x,&y,&z);
   t2=(float)clock();

   dt=(t2-t1)/(float)(CLOCKS_PER_SEC);
   dt*=1.0e6f/(float)(4*NLOOP);

   printf("Average time per value of x: %4.3f micro sec\n\n",dt);
   exit(0);
}

#else

int main(void)
{
   printf("\n");
   printf("Timing of the SSE code for ln(1+x), sin(pi*x) and cos(pi*x)\n");
   printf("-----------------------------------------------------------\n\n");

   printf("SSE needs to be switched on at compilation time\n");
   printf("(flags SSE, SSE2 or SSE3)\n\n");

   exit(0);
}

#endif
