
/*******************************************************************************
*
* File time6.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of cmat_vec and cmat_mul
*
*******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include "random.h"
#include "su3.h"
#include "start.h"
#include "misc.h"


static void rvec(int n,complex *v)
{
   gauss((float*)(v),2*n);
}


static void rmat(int n,complex *a)
{
   int i,j;
   float r;

   r=0.1f/(float)(n*n);

   gauss((float*)(a),2*n*n);

   for (i=0;i<n;i++)
   {
      for (j=0;j<n;j++)
      {
         a[i*n+j].re*=r;
         a[i*n+j].im*=r;
      }

      a[i*n+i].re+=1.0f;
   }
}


int main(void)
{
   int n,nm,nsq,nr;
   int i,count;
   float t1,t2,dt;
   complex *a,*b,*c,*v,*w;

   printf("\n");
   printf("Timing of cmat_vec and cmat_mul\n");
   printf("-------------------------------\n\n");

   printf("Matrix size: ");
   scanf(" %d",&n);

   error(n<1,1,"main [time6.c]","Improper matrix size");

   nm=(int)(2.0e7f/(float)(n*n*sizeof(complex)))+1;
   a=amalloc(nm*n*(3*n+2)*sizeof(*a),4);
   error(a==NULL,1,"main [time6.c]","Unable to allocate auxiliary arrays");

   nsq=n*n;
   b=a+nm*nsq;
   c=b+nm*nsq;
   v=c+nm*nsq;
   w=v+nm*n;
   
   for (i=0;i<nm;i++)
   {
      rmat(n,a+i*nsq);
      rmat(n,b+i*nsq);
      rmat(n,c+i*nsq);
      rvec(n,v+i*n);
      rvec(n,w+i*n);      
   }

   nr=(int)(2.0e9f/(float)(nm*nsq*4)+1);

   printf("\n");
   printf("The times per application are:\n\n");
   
   t1=(float)clock();
   for (count=0;count<nr;count++)
   {
      for (i=0;i<nm;i++)
         cmat_vec(n,a+i*nsq,v+i*n,w+i*n);
   }
   t2=(float)clock();

   dt=(t2-t1)/(float)(CLOCKS_PER_SEC);
   dt*=1.0e6/(float)(nr*nm);

   printf("cmat_vec:   %4.3f micro sec (%d Mflops)\n",
          dt,(int)((float)(n*((n-1)*8+6))/dt));

   nr=(int)(2.0e9f/(float)(nm*nsq*n*4)+1);
   
   t1=(float)clock();
   for (count=0;count<nr;count++)
   {
      for (i=0;i<nm;i++)
         cmat_mul(n,a+i*nsq,b+i*nsq,c+i*nsq);
   }
   t2=(float)clock();

   dt=(t2-t1)/(float)(CLOCKS_PER_SEC);
   dt*=1.0e6f/(float)(nr*nm);

   printf("cmat_mul:   %4.3f micro sec (%d Mflops)\n\n",
          dt,(int)((float)(n*n*((n-1)*8+6))/dt));   

   exit(0);
}
