
/*******************************************************************************
*
* File check3.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the inversion program sap_gcr
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "update.h"
#include "linalg.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,bc,nth,bs[4];
   int id0,id1,id2;
   int n,nkv,nmr,ncy,nmx,status,ifail;
   double beta,kappa,csw,res,res1,res2;
   double wt1,wt2,wdt;
   complex_dble z;
   spinor_dble **wsd,*pd0,*pd1,*pd2;
   lat_parms_t lat;
   gcr_parms_t gcr;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check3.log","w",stdout);
      fin=freopen("check3.in","r",stdin);

      printf("\n");
      printf("Check of the inversion program sap_gcr\n");
      printf("--------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("beta","%lf\n",&beta);
      read_line("kappa","%lf\n",&kappa);
      read_line("csw","%lf\n",&csw);      
      read_line("bc","%d\n",&bc);
      read_line("nth","%d\n",&nth);
      read_line("bs","%d %d %d %d\n",&bs[0],&bs[1],&bs[2],&bs[3]);
      read_line("nmr","%d\n",&nmr);
      read_line("ncy","%d\n",&ncy);
      read_line("nkv","%d\n",&nkv);      
      read_line("nmx","%d\n",&nmx);
      read_line("res","%lf",&res);
      fclose(fin);

      printf("beta = %.2f\n",beta);
      printf("kappa = %.4f\n",kappa);
      printf("csw = %.4f\n",csw);
      printf("bc = %d\n",bc);
      printf("nth = %d\n",nth);
      printf("bs = %d %d %d %d\n\n",bs[0],bs[1],bs[2],bs[3]);
      printf("nmr = %d\n",nmr);
      printf("ncy = %d\n",ncy);
      printf("nkv = %d\n",nkv);
      printf("nmx = %d\n",nmx);
      printf("res = %.2e\n\n",res);

      error_root(nkv>40,1,"main [check3.c]",
                 "Number nkv of Krylov vectors is greater than 40");
      fflush(flog);
   }

   MPI_Bcast(&beta,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&kappa,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&csw,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nth,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(bs,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmr,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&ncy,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nkv,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&nmx,1,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&res,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

   start_ranlux(0,12345);
   geometry();
   alloc_u();
   alloc_ud();
   alloc_swd();

   alloc_s(2*nkv+1);
   alloc_sd(4);
   wsd=reserve_wsd(3);

   set_lat_parms(beta,kappa,csw);
   set_gcr_parms(bs,nmr,ncy,nkv);
   lat=lat_parms();
   gcr=gcr_parms();

   random_u();

   for (n=0;n<nth;n++)
      update((float)(beta),1,4,10);

   assign_u2ud();
   if (bc==-1)
      flipbcd();

   pd0=wsd[0];
   pd1=wsd[1];
   pd2=wsd[2];

   id0=(pd0-psd[0][0])/NSPIN;
   id1=(pd1-psd[0][0])/NSPIN;
   id2=(pd2-psd[0][0])/NSPIN;

   random_sd(VOLUME,pd0,1.0);
   normalize_dble(VOLUME,1,pd0);

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();
   res1=sap_gcr(nmx,res,id0,id1,&status,&ifail);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();
   wdt=wt2-wt1;

   print_flags();
   print_blk_flags(GCR_BLOCKS,0);

   z.re=-1.0;
   z.im=0.0;
   sw_term();
   Qnohat_dble(id1,id2);
   mulg5_dble(VOLUME,pd2);
   mulc_spinor_add_dble(VOLUME,pd0,pd2,z);
   res2=norm_square_dble(VOLUME,1,pd0);
   res2=sqrt(res2);
   error_chk();
   release_wsd();

   if (my_rank==0)
   {
      printf("status = %d, ifail = %d\n",status,ifail);
      printf("res = %.2e, check = %.2e, time = %.2e sec\n\n",res1,res2,wdt);
      fclose(flog);
   }

   MPI_Finalize();
   exit(0);
}
