
/*******************************************************************************
*
* File check1.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check that the "staples" match the change in the plaquette action if
* a single link is rotated
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"


static float plaq(int ix,int mu,int nu)
{
   int iy,iz;
   su3 *v1,*v2,*v3,*v4,w1,w2,w3;

   iy=iup[ix][mu];
   iz=iup[ix][nu];

   v1=pu[ix][mu];
   v2=pu[iy][nu];
   v3=pu[iz][mu];
   v4=pu[ix][nu];

   _su3_times_su3(w1,(*v1),(*v2));
   _su3_times_su3(w2,(*v4),(*v3));
   _su3_dagger(w3,w2);
   _su3_times_su3(w2,w1,w3);      
      
   return w2.c11.re+w2.c22.re+w2.c33.re;
}


static float all_plaq(int ix,int mu)
{
   int nu,iy;
   float s;

   s=0.0f;

   for (nu=0;nu<4;nu++)
   {
      if (nu!=mu)
      {
         iy=idn[ix][nu];
         s+=plaq(ix,mu,nu);
         s+=plaq(iy,mu,nu);
      }
   }
   
   return s;
}


int main(int argc,char *argv[])
{
   int my_rank,ix,mu,n,nlks;
   double d,dmax,dmax_all;
   su3 u,v,vdag,r,w;
   link_t *lks;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check1.log","w",stdout);
   
      printf("\n");
      printf("Check of the program 'staples'\n");
      printf("------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }   
   
   start_ranlux(0,12345);
   geometry();
   alloc_u();   
   random_u();
   lks=active_links(&nlks);
   
   dmax=0.0;

   for (n=0;n<nlks;n++)
   {
      ix=lks[n].ix;
      mu=lks[n].mu;
      
      staples(ix,mu,&v);
      _su3_dagger(vdag,v);
      random_su3(&r);
      _su3_times_su3(w,r,vdag);

      u=*(pu[ix][mu]);         
      *(pu[ix][mu])=r; 

      d=(double)(all_plaq(ix,mu))-(double)(w.c11.re+w.c22.re+w.c33.re);
      d=fabs(d);
      if (d>dmax)
         dmax=d;

      *(pu[ix][mu])=u;
   }

   MPI_Reduce(&dmax,&dmax_all,1,MPI_DOUBLE,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {   
      printf("Maximal absolute deviation from the action: %.1e\n",dmax_all);
      printf("(Should be around 1.0e-6)\n\n");
      fclose(flog);
   }   

   MPI_Finalize();    
   exit(0);
}
