
/*******************************************************************************
*
* File check6.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the array of the active links
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"

static int ilks[VOLUME][4];


static int btest(int ix)
{
   int iy,iz,mu;

   if (ix>=VOLUME)
      return 1;
   
   for (mu=0;mu<4;mu++)
   {
      iy=iup[ix][mu];
      iz=idn[ix][mu];

      if ((iy>=VOLUME)||(iz>=VOLUME))
         return 1;
   }
   
   return 0;
}


int main(int argc,char *argv[])
{
   int my_rank,ix,iy,mu,n,nlks;
   int it[4],itall[4];
   link_t *lks;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check6.log","w",stdout);
   
      printf("\n");
      printf("Check of the array of the active links\n");
      printf("--------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }   
   
   start_ranlux(0,12345);
   geometry();
   lks=active_links(&nlks);

   it[0]=0;
   it[1]=0;
   it[2]=0;
   it[3]=0;
   
   for (ix=0;ix<VOLUME;ix++)
   {
      for (mu=0;mu<4;mu++)
      {
         iy=iup[ix][mu];

         if ((btest(ix)==1)&&(btest(iy)==1))
            ilks[ix][mu]=2;
         else
            ilks[ix][mu]=0;
      }
   }

   for (n=0;n<nlks;n++)
   {
      ix=lks[n].ix;
      mu=lks[n].mu;

      if ((ix<0)||(ix>=VOLUME)||(mu<0)||(mu>=4))
         it[0]=1;
      else
      {
         iy=iup[ix][mu];

         if ((btest(ix)==1)&&(btest(iy)==1))
            it[1]=1;
         
         if (ilks[ix][mu]==1)
            it[2]=1;

         ilks[ix][mu]=1;
      }
   }

   for (ix=0;ix<VOLUME;ix++)
   {
      for (mu=0;mu<4;mu++)
      {
         if (ilks[ix][mu]==0)
            it[3]=1;
      }
   }
   
   MPI_Reduce(it,itall,4,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);
 
   if (my_rank==0)
   {
      if (itall[0]==1)
         printf("Array elements are out of range\n");
      if (itall[1]==1)
         printf("Some active links are not inside the local lattice\n");
      if (itall[2]==1)
         printf("Some active links are double-counted\n");
      if (itall[3]==1)
         printf("Some active links have been omitted\n");

      if ((itall[0]==0)&&(itall[1]==0)&&(itall[2]==0)&&(itall[3]==0))
         printf("No errors discovered\n");

      printf("\n");
      fclose(flog);
   }   

   MPI_Finalize();    
   exit(0);
}
