
/*******************************************************************************
*
* File time2.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the cabmar program
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"

#define N 1000000

#if ((defined SSE)||(defined SSE2))
static su3 u,v __attribute__ ((aligned (8)));
#else
static su3 u,v;
#endif


static void random_staples(su3 *v)
{
   int i;
   float *r;

   r=(float*)(v);
   gauss(r,18);

   for (i=0;i<18;i++)
      r[i]*=3.0f;
}


int main(int argc,char *argv[])
{
   int my_rank,count;
   float beta;
   double wt1,wt2,wdt;     
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      flog=freopen("time2.log","w",stdout);
   
      printf("\n");
      printf("Timing of the Cabibbo-Marinari rotations\n");
      printf("----------------------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);      
      fflush(flog);
   }

   beta=6.0f;   
   start_ranlux(0,123456);   
   random_su3(&u);
   random_staples(&v);

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();   
   for (count=0;count<N;count++)
      cabmar(beta,&u,&v,0);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(N);

   if (my_rank==0)
      printf("Time per link (heatbath at beta=6.0): %4.3f micro sec\n\n",wdt);
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();  
   for (count=0;count<N;count++)
      cabmar(beta,&u,&v,1);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();   

   wdt=1.0e6*(wt2-wt1)/(double)(N);   

   if (my_rank==0)
   {
      printf("Time per link (over-relaxation): %4.3f micro sec\n\n",wdt);
      fclose(flog);
   }   

   MPI_Finalize();  
   exit(0);
}
