
/*******************************************************************************
*
* File time3.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Timing of the update program
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "update.h"
#include "global.h"


int main(int argc,char *argv[])
{
   int my_rank,n,count,nlks;
   float beta;
   double wt1,wt2,wdt;
   link_t *lks;
   FILE *flog=NULL;   

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);
   
   if (my_rank==0)
   {
      flog=freopen("time3.log","w",stdout);

      printf("\n");
      printf("Timing of the update program\n");
      printf("----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);
      fflush(flog);
   }

   beta=6.0f;
   start_ranlux(0,12345);
   geometry();
   alloc_u();
   random_u();
   
   lks=active_links(&nlks);
   afree(lks);
   
   n=100000/(int)(nlks);
   if (n<2)
      n=2;   
   
   for (count=0;count<10;count++)
      update(beta,1,2,3);

   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime();    
   for (count=0;count<n;count++)
      update(beta,10,0,10);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();  

   wdt=1.0e6*(wt2-wt1)/(double)(10*n*nlks);

   if (my_rank==0)
   {   
      printf("beta = %.1f, number of active links = %d\n\n",beta,nlks);
      printf("Time per link update (heatbath): %4.3f micro sec\n",wdt);
   }
   
   MPI_Barrier(MPI_COMM_WORLD);
   wt1=MPI_Wtime(); 
   for (count=0;count<n;count++)
      update(beta,0,10,10);
   MPI_Barrier(MPI_COMM_WORLD);
   wt2=MPI_Wtime();  

   wdt=1.0e6*(wt2-wt1)/(double)(10*n*nlks);

   if (my_rank==0)
   {     
      printf("Time per link update (over-relaxation): %4.3f micro sec\n\n",wdt);
      fclose(flog);
   }   

   MPI_Finalize();
   exit(0);
}
