
/*******************************************************************************
*
* File cabmar.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Cabibbo-Marinari rotations
*
* The externally accessible function is
*
*   void cabmar(float beta,su3 *u, su3 *v,int type)
*     Updates the SU(3) matrix u with probability density proportional to
*     exp{(beta/3)*Re[tr(u*v^dagger)]} by applying three Cabibbo-Marinari 
*     heatbath (if type=0) or over-relaxation (if type=1) rotations
*
*******************************************************************************/

#define CABMAR_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "su3.h"
#include "random.h"


#if (defined SSE)

#include "sse.h"

static float s[4],w[4] __attribute__ ((aligned (16)));
static su3_vector *pu1,*pu2,*pv1,*pv2;


static void rotate(void)
{
   __asm__ __volatile__ ("movss %0, %%xmm0 \n\t"
                         "movss %1, %%xmm1 \n\t"
                         "movss %2, %%xmm2 \n\t"
                         "movss %3, %%xmm3 \n\t"
                         "movlps %7, %%xmm7 \n\t"
                         "movhps %8, %%xmm7 \n\t"
                         "shufps $0x0, %%xmm0, %%xmm0 \n\t"
                         "shufps $0x0, %%xmm1, %%xmm1 \n\t"
                         "shufps $0x0, %%xmm2, %%xmm2 \n\t"
                         "shufps $0x0, %%xmm3, %%xmm3 \n\t"
                         "mulps %4, %%xmm1 \n\t"
                         "mulps %5, %%xmm2 \n\t"
                         "mulps %6, %%xmm3"
                         :
                         :
                         "m" (s[0]),
                         "m" (s[1]),
                         "m" (s[2]),
                         "m" (s[3]),
                         "m" (_sse_sgn24),
                         "m" (_sse_sgn12),
                         "m" (_sse_sgn23),
                         "m" ((*pu1).c1),
                         "m" ((*pu2).c1)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3", "xmm7");
   
   __asm__ __volatile__ ("movaps %%xmm1, %%xmm4 \n\t"
                         "movaps %%xmm2, %%xmm5 \n\t"
                         "movaps %%xmm3, %%xmm6 \n\t"
                         "mulps %%xmm7, %%xmm4 \n\t"
                         "mulps %%xmm7, %%xmm5 \n\t"
                         "shufps $0x1b, %%xmm4, %%xmm4 \n\t"
                         "mulps %%xmm7, %%xmm6 \n\t"
                         "shufps $0x4e, %%xmm5, %%xmm5 \n\t"
                         "shufps $0xb1, %%xmm6, %%xmm6 \n\t"
                         "mulps %%xmm0, %%xmm7 \n\t"
                         "addps %%xmm4, %%xmm5 \n\t"
                         "addps %%xmm7, %%xmm6"
                         :
                         :
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("movlps %2, %%xmm7 \n\t"
                         "movhps %3, %%xmm7 \n\t"
                         "movaps %%xmm1, %%xmm4 \n\t"                         
                         "addps %%xmm5, %%xmm6 \n\t"
                         "mulps %%xmm7, %%xmm4 \n\t"
                         "movaps %%xmm2, %%xmm5 \n\t"
                         "movlps %%xmm6, %0 \n\t"
                         "movhps %%xmm6, %1 \n\t"
                         "mulps %%xmm7, %%xmm5 \n\t"                         
                         "movaps %%xmm3, %%xmm6 \n\t"
                         "shufps $0x1b, %%xmm4, %%xmm4 \n\t"
                         "mulps %%xmm7, %%xmm6 \n\t"
                         "shufps $0x4e, %%xmm5, %%xmm5 \n\t"
                         "mulps %%xmm0, %%xmm7 \n\t"
                         "shufps $0xb1, %%xmm6, %%xmm6 \n\t"
                         "addps %%xmm4, %%xmm5 \n\t"
                         "addps %%xmm7, %%xmm6"
                         :
                         "=m" ((*pu1).c1),
                         "=m" ((*pu2).c1)
                         :
                         "m" ((*pu1).c2),
                         "m" ((*pu2).c2)
                         :
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("movlps %4, %%xmm4 \n\t"
                         "movhps %5, %%xmm4 \n\t"
                         "addps %%xmm5, %%xmm6 \n\t"
                         "mulps %%xmm4, %%xmm1 \n\t"
                         "mulps %%xmm4, %%xmm2 \n\t"
                         "shufps $0x1b, %%xmm1, %%xmm1 \n\t"
                         "mulps %%xmm4, %%xmm3 \n\t"
                         "shufps $0x4e, %%xmm2, %%xmm2 \n\t"
                         "mulps %%xmm4, %%xmm0 \n\t"
                         "shufps $0xb1, %%xmm3, %%xmm3 \n\t"
                         "addps %%xmm1, %%xmm2 \n\t"
                         "addps %%xmm0, %%xmm3 \n\t"
                         "addps %%xmm2, %%xmm3 \n\t"
                         "movlps %%xmm6, %0 \n\t"
                         "movhps %%xmm6, %1 \n\t"                         
                         "movlps %%xmm3, %2 \n\t"
                         "movhps %%xmm3, %3"
                         :
                         "=m" ((*pu1).c2),
                         "=m" ((*pu2).c2),
                         "=m" ((*pu1).c3),
                         "=m" ((*pu2).c3)
                         :
                         "m" ((*pu1).c3),
                         "m" ((*pu2).c3)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm6");
}


static void wmatrix(void)
{
   __asm__ __volatile__ ("movlps %0, %%xmm0 \n\t"
                         "movlps %1, %%xmm1 \n\t"
                         "movlps %2, %%xmm2 \n\t"
                         "movhps %3, %%xmm0 \n\t"
                         "movhps %4, %%xmm1 \n\t"
                         "movhps %5, %%xmm2"
                         :
                         :
                         "m" ((*pu1).c1),
                         "m" ((*pu1).c2),
                         "m" ((*pu1).c3),
                         "m" ((*pu2).c1),
                         "m" ((*pu2).c2),
                         "m" ((*pu2).c3)
                         :
                         "xmm0", "xmm1", "xmm2");   
   
   __asm__ __volatile__ ("movss %0, %%xmm3 \n\t"
                         "movss %1, %%xmm4 \n\t"
                         "movss %2, %%xmm5 \n\t"
                         "movss %3, %%xmm6 \n\t"
                         "movss %4, %%xmm7 \n\t"
                         "shufps $0x0, %%xmm3, %%xmm3 \n\t"
                         "shufps $0x0, %%xmm4, %%xmm4 \n\t"
                         "shufps $0x0, %%xmm5, %%xmm5 \n\t"
                         "shufps $0x0, %%xmm6, %%xmm6 \n\t"
                         "shufps $0x0, %%xmm7, %%xmm7 \n\t"
                         "mulps %%xmm0, %%xmm3 \n\t"
                         "mulps %%xmm1, %%xmm4 \n\t"
                         "mulps %%xmm2, %%xmm5 \n\t"                         
                         "mulps %%xmm0, %%xmm6 \n\t"
                         "mulps %%xmm1, %%xmm7 \n\t"
                         "addps %%xmm5, %%xmm3 \n\t"
                         "addps %%xmm6, %%xmm4 \n\t"
                         "addps %%xmm7, %%xmm3"
                         :
                         :
                         "m" ((*pv1).c1.re),
                         "m" ((*pv2).c2.re),
                         "m" ((*pv1).c3.re),
                         "m" ((*pv2).c1.re),
                         "m" ((*pv1).c2.re)
                         :
                         "xmm3", "xmm4", "xmm5", "xmm6", "xmm7");
   
   __asm__ __volatile__ ("movss %0, %%xmm5 \n\t"
                         "movss %1, %%xmm6 \n\t"
                         "movss %2, %%xmm7 \n\t"
                         "shufps $0x0, %%xmm5, %%xmm5 \n\t"
                         "shufps $0x0, %%xmm6, %%xmm6 \n\t"
                         "shufps $0x0, %%xmm7, %%xmm7 \n\t"
                         "shufps $0xb1, %%xmm0, %%xmm0 \n\t"
                         "mulps %%xmm2, %%xmm5 \n\t"                         
                         "shufps $0xb1, %%xmm1, %%xmm1 \n\t"
                         "shufps $0xb1, %%xmm2, %%xmm2 \n\t"
                         "mulps %3, %%xmm0 \n\t"                         
                         "mulps %3, %%xmm1 \n\t"
                         "mulps %3, %%xmm2 \n\t"
                         "mulps %%xmm0, %%xmm6 \n\t"
                         "addps %%xmm5, %%xmm4 \n\t"
                         "mulps %%xmm0, %%xmm7 \n\t"
                         "addps %%xmm6, %%xmm3 \n\t"
                         "addps %%xmm7, %%xmm4"
                         :
                         :
                         "m" ((*pv2).c3.re),
                         "m" ((*pv1).c1.im),
                         "m" ((*pv2).c1.im),
                         "m" (_sse_sgn24)
                         :
                         "xmm0", "xmm1", "xmm2", "xmm3",
                         "xmm4", "xmm5", "xmm6", "xmm7");

   __asm__ __volatile__ ("movss %0, %%xmm0 \n\t"
                         "movss %1, %%xmm5 \n\t"
                         "movss %2, %%xmm6 \n\t"
                         "movss %3, %%xmm7 \n\t"
                         "shufps $0x0, %%xmm0, %%xmm0 \n\t"
                         "shufps $0x0, %%xmm5, %%xmm5 \n\t"
                         "shufps $0x0, %%xmm6, %%xmm6 \n\t"
                         "shufps $0x0, %%xmm7, %%xmm7 \n\t"
                         "mulps %%xmm1, %%xmm0 \n\t"
                         "mulps %%xmm1, %%xmm5 \n\t"
                         "addps %%xmm0, %%xmm3 \n\t"
                         "mulps %%xmm2, %%xmm6 \n\t"
                         "addps %%xmm5, %%xmm4 \n\t"
                         "mulps %%xmm2, %%xmm7 \n\t"
                         "addps %%xmm6, %%xmm3 \n\t"
                         "addps %%xmm7, %%xmm4"                         
                         :
                         :
                         "m" ((*pv1).c2.im),
                         "m" ((*pv2).c2.im),
                         "m" ((*pv1).c3.im),
                         "m" ((*pv2).c3.im)
                         :
                         "xmm0", "xmm3", "xmm4", 
                         "xmm5", "xmm6", "xmm7"); 

   __asm__ __volatile__ ("shufps $0x6c, %%xmm3, %%xmm3 \n\t"
                         "shufps $0xc6, %%xmm4, %%xmm4 \n\t"
                         "mulps %1, %%xmm3 \n\t"
                         "mulps %2, %%xmm4 \n\t"
                         "addps %%xmm3, %%xmm4 \n\t"
                         "movaps %%xmm4, %0"
                         :
                         "=m" (w)
                         :
                         "m" (_sse_sgn24),
                         "m" (_sse_sgn23)
                         :
                         "xmm3", "xmm4");
}

#else

static float s[4],w[4];
static su3_vector *pu1,*pu2,*pv1,*pv2;


static void rotate(void)
{
   complex z1,z2;

   z1.re=
      s[0]*(*pu1).c1.re-s[1]*(*pu2).c1.im+s[2]*(*pu2).c1.re-s[3]*(*pu1).c1.im;
   z1.im=
      s[0]*(*pu1).c1.im+s[1]*(*pu2).c1.re+s[2]*(*pu2).c1.im+s[3]*(*pu1).c1.re;
   z2.re=
      s[0]*(*pu2).c1.re-s[1]*(*pu1).c1.im-s[2]*(*pu1).c1.re+s[3]*(*pu2).c1.im;
   z2.im=
      s[0]*(*pu2).c1.im+s[1]*(*pu1).c1.re-s[2]*(*pu1).c1.im-s[3]*(*pu2).c1.re;
   (*pu1).c1=z1;
   (*pu2).c1=z2;

   z1.re=
      s[0]*(*pu1).c2.re-s[1]*(*pu2).c2.im+s[2]*(*pu2).c2.re-s[3]*(*pu1).c2.im;
   z1.im=
      s[0]*(*pu1).c2.im+s[1]*(*pu2).c2.re+s[2]*(*pu2).c2.im+s[3]*(*pu1).c2.re;
   z2.re=
      s[0]*(*pu2).c2.re-s[1]*(*pu1).c2.im-s[2]*(*pu1).c2.re+s[3]*(*pu2).c2.im;
   z2.im=
      s[0]*(*pu2).c2.im+s[1]*(*pu1).c2.re-s[2]*(*pu1).c2.im-s[3]*(*pu2).c2.re;
   (*pu1).c2=z1;
   (*pu2).c2=z2;

   z1.re=
      s[0]*(*pu1).c3.re-s[1]*(*pu2).c3.im+s[2]*(*pu2).c3.re-s[3]*(*pu1).c3.im;
   z1.im=
      s[0]*(*pu1).c3.im+s[1]*(*pu2).c3.re+s[2]*(*pu2).c3.im+s[3]*(*pu1).c3.re;
   z2.re=
      s[0]*(*pu2).c3.re-s[1]*(*pu1).c3.im-s[2]*(*pu1).c3.re+s[3]*(*pu2).c3.im;
   z2.im=
      s[0]*(*pu2).c3.im+s[1]*(*pu1).c3.re-s[2]*(*pu1).c3.im-s[3]*(*pu2).c3.re;
   (*pu1).c3=z1;
   (*pu2).c3=z2;   
}


static void wmatrix(void)
{
   w[0]=
      (*pu2).c1.re*(*pv2).c1.re+(*pu2).c1.im*(*pv2).c1.im+
      (*pu2).c2.re*(*pv2).c2.re+(*pu2).c2.im*(*pv2).c2.im+
      (*pu2).c3.re*(*pv2).c3.re+(*pu2).c3.im*(*pv2).c3.im+
      (*pu1).c1.re*(*pv1).c1.re+(*pu1).c1.im*(*pv1).c1.im+
      (*pu1).c2.re*(*pv1).c2.re+(*pu1).c2.im*(*pv1).c2.im+
      (*pu1).c3.re*(*pv1).c3.re+(*pu1).c3.im*(*pv1).c3.im;

   w[1]=
      (*pu2).c1.re*(*pv1).c1.im-(*pu2).c1.im*(*pv1).c1.re+
      (*pu2).c2.re*(*pv1).c2.im-(*pu2).c2.im*(*pv1).c2.re+
      (*pu2).c3.re*(*pv1).c3.im-(*pu2).c3.im*(*pv1).c3.re+
      (*pu1).c1.re*(*pv2).c1.im-(*pu1).c1.im*(*pv2).c1.re+
      (*pu1).c2.re*(*pv2).c2.im-(*pu1).c2.im*(*pv2).c2.re+
      (*pu1).c3.re*(*pv2).c3.im-(*pu1).c3.im*(*pv2).c3.re;   

   w[2]=
      (*pu2).c1.re*(*pv1).c1.re+(*pu2).c1.im*(*pv1).c1.im+
      (*pu2).c2.re*(*pv1).c2.re+(*pu2).c2.im*(*pv1).c2.im+
      (*pu2).c3.re*(*pv1).c3.re+(*pu2).c3.im*(*pv1).c3.im-
      (*pu1).c1.re*(*pv2).c1.re-(*pu1).c1.im*(*pv2).c1.im-
      (*pu1).c2.re*(*pv2).c2.re-(*pu1).c2.im*(*pv2).c2.im-
      (*pu1).c3.re*(*pv2).c3.re-(*pu1).c3.im*(*pv2).c3.im;

   w[3]=
      (*pu1).c1.re*(*pv1).c1.im-(*pu1).c1.im*(*pv1).c1.re+
      (*pu1).c2.re*(*pv1).c2.im-(*pu1).c2.im*(*pv1).c2.re+
      (*pu1).c3.re*(*pv1).c3.im-(*pu1).c3.im*(*pv1).c3.re-
      (*pu2).c1.re*(*pv2).c1.im+(*pu2).c1.im*(*pv2).c1.re-
      (*pu2).c2.re*(*pv2).c2.im+(*pu2).c2.im*(*pv2).c2.re-
      (*pu2).c3.re*(*pv2).c3.im+(*pu2).c3.im*(*pv2).c3.re;
}

#endif

void cabmar(float beta,su3 *u,su3 *v,int type)
{
   int i;
   float b,bsq,wsq,rho,fact,r[4];

   pu1=(su3_vector*)(u);
   pu2=pu1+1;
   pv1=(su3_vector*)(v);
   pv2=pv1+1;

   b=0.33333333f*beta;
   bsq=b*b;
   
   for (i=0;i<3;i++)
   {
      wmatrix();
      wsq=w[0]*w[0]+w[1]*w[1]+w[2]*w[2]+w[3]*w[3];

      if ((bsq*wsq)>1.0e-16f)
      {
         if (type==1)
         {
            fact=(w[0]+w[0])/wsq;
            s[0]=fact*w[0]-1.0f;
            s[1]=fact*w[1];
            s[2]=fact*w[2];
            s[3]=fact*w[3];         
         }
         else
         {
            fact=(float)(sqrt((double)(wsq)));
            rho=b*fact;

            random_su2(rho,r);

            fact=1.0f/fact;            
            s[0]=fact*(r[0]*w[0]-r[1]*w[1]-r[2]*w[2]-r[3]*w[3]);
            s[1]=fact*(r[1]*w[0]+r[0]*w[1]-r[2]*w[3]+r[3]*w[2]);
            s[2]=fact*(r[2]*w[0]+r[0]*w[2]-r[3]*w[1]+r[1]*w[3]);
            s[3]=fact*(r[3]*w[0]+r[0]*w[3]-r[1]*w[2]+r[2]*w[1]);         
         }      
      }
      else
         random_su2(0.0f,s);

      rotate();

      if (i==0)
      {
         pu2=pu2+1;
         pv2=pv2+1;
      }
      else if (i==1)
      {
         pu1=pu1+1;
         pv1=pv1+1;
      }
   }
}

